#include "portable.h"

#include <stdio.h>

#include <ac/string.h>
#include <ac/socket.h>

#include "ldap_pvt.h"
#include "lutil.h"
#include "slap.h"
#include "lutil_ldap.h"

#include <lber.h>
#include <ldif.h>
#include <lutil.h>
#ifdef HAVE_TLS
#include <openssl/x509.h>
#include <openssl/err.h>
#include <openssl/rsa.h>
#include <openssl/crypto.h>
#include <openssl/pem.h>
#include <openssl/bio.h>
#include <openssl/asn1.h>
#include <openssl/x509v3.h>
#include <openssl/ssl.h>
#include <openssl/md5.h>
#endif

#include "xps.h"

#define LDAPMOD_MAXLINE 4096

OidFunction LookupTable[]={
{&decodeC, "2.5.4.6", "C"},
{&decodeO, "2.5.4.10", "O"},
{&decodeOu, "2.5.4.11", "Ou"},
{&decodeCn, "2.5.4.3", "Cn"},
{&decodeL, "2.5.4.7", "L"},
{&decodeSt, "2.5.4.8", "St"},
{&decodeE_mail, "2.5.4.99", "E_mail"},
{&decodeDc, "0.9.2342.19200300.100.1.25", "Dc"},
{&decodeAuthorityKeyIdentifier, "2.5.29.35", "AuthorityKeyIdentifier"},
{&decodeSubjectKeyIdentifier, "2.5.29.14", "SubjectKeyIdentifier"},
{&decodeKeyUsage, "2.5.29.15", "KeyUsage"},
{&decodePrivateKeyUsagePeriod, "2.5.29.16", "PrivateKeyUsagePeriod"},
{&decodePolicyMappings, "2.5.29.33", "PolicyMappings"},
{&decodeSubjectAltName, "2.5.29.17", "SubjectAltName"},
{&decodeIssuerAltName, "2.5.29.18", "IssuerAltName"},
{&decodeSubjectDirectoryAttributes, "2.5.29.9", "SubjectDirectoryAttributes"},
{&decodeBasicConstraints, "2.5.29.19", "BasicConstraints"},
{&decodeNameConstraints, "2.5.29.30", "NameConstraints"},
{&decodePolicyConstraints, "2.5.29.36", "PolicyConstraints"},
{&decodeCRLDistributionPoints, "2.5.29.31", "CRLDistributionPoints"},
{&decodeExtKeyUsageSyntax, "2.5.29.37", "ExtKeyUsageSyntax"},
{&decodeInhibitAnyPolicy, "2.5.29.54", "InhibitAnyPolicy"},
{&decodeFreshestCRL, "2.5.29.46", "FreshestCRL"},
{&decodeAuthorityInfoAccessSyntax, "1.3.6.1.5.5.7.1.1", "AuthorityInfoAccessSyntax"},
{&decodeSubjectInfoAccessSyntax, "1.3.6.1.5.5.7.1.11", "SubjectInfoAccessSyntax"},
{&decodeCRLNumber, "2.5.29.20", "CRLNumber"},
{&decodeIssuingDistributionPoint, "2.5.29.28", "IssuingDistributionPoint"},
{&decodeBaseCRLNumber, "2.5.29.27", "BaseCRLNumber"},
{&decodeCRLReason, "2.5.29.21", "CRLReason"},
{&decodeCertificateIssuer, "2.5.29.29", "CertificateIssuer"},
{&decodeHoldInstructionCode, "2.5.29.23", "HoldInstructionCode"},
{&decodeInvalidityDate, "2.5.29.24", "InvalidityDate"},
{NULL, NULL, NULL}
};

/* the asn1 decoder library (c) E Ball 2002-2003 */

/* _CPFPLUS functions return empty string in case of success 
	and NULL in case of an error */

OID lastOID = { 0, NULL };

char *Primitive[] = {
	"",
	"BOOLEAN",
	"INTEGER",
	"BIT STRING",
	"OCTET STRING",
	"NULL",
	"OBJECT IDENTIFIER",
	"",
	"",
	"",
	"ENUMERATED",
	"",
	"",
	"",
	"",
	"",
	"SEQUENCE",
	"SET",
	"NUMERIC STRING",
	"T61 STRING",
	"IA5 STRING",
	"UTCTIME",
	"GENERALISED TIME",
	"GRAPHICAL STRING",
	"VISIBLE STRING",
	"UNIVERSAL STRING",
	"BMP STRING"
};

char *intToChar( long value )
{
	int i, tmp_val = value;
	char *tmp;
	
	if ( tmp_val == 0 )
		i = 1;
	else {
		i = 0;
		while ( tmp_val > 0 )
		{
			tmp_val /= 10;
			i++;
		}
	}
	
	tmp = ( char *)malloc( sizeof( char ) * ( i + 1 ) );
	sprintf (tmp, "%ld", ( long )value);
	return tmp;
}

void xps_cleanup()
{
	if ( lastOID.oidbytes != NULL )
		free( lastOID.oidbytes );
}

//returns the length of the BER block
int
decodeTL( byte **PP )
{
	byte *T, b;
	int tag, L, noOfBytes;

	T = *PP;
	tag = *T & 0x1f;
	// dont handle multibyte tags for the moment
	if ( tag == 31 )
	{
		//printf ("multi-byte tag not supported yet\n");
		return (-1);
	}
	
	b = *( ++T );  // move on to length field
	if ( b & 0x80 ) // multi-byte length
	{
		int i;
		noOfBytes = (byte) * T++ & 0x7f;
		for ( L = 0, i = 0; i < noOfBytes; i++ )
		{
			L = L * 256 + (byte) * T++;
		}
	} else {	// single byte length
		L = b & 0x7f;
		T++;
	}
	*PP = T;
	return L;
}

// this routine sees if a user extra tag is there & takes it out
// either returns the tag or -1 if the item is optional but not present
// if it is a tagged type tag is the expected tag 
// if there is no tag - tag is -1
// return 0 if a tag was expected but not found
int
decodeTag (int tag, int implicit, byte ** P, int optional, int Default)
{
	byte *T;
	int tlvTag, L;

	T = *P;
	tlvTag = *T;

	if (tag == -1 && optional)	// -1 means we do not anticpate any particular tag e.g ANY DFB
		return tlvTag;

	/* TODO:  check this statement - it causes error when parsing "faultyCert.der" */
	if ( ( tlvTag ) != tag && ( optional || Default ) ) {
		return -1;
	}

	if ( ( tlvTag & 0xc0 ) == 0x80 )	// a context tag
	{
		if ( tlvTag != tag && tag != -1 )
		{
			if ( !( optional || Default ) )
			{
//				printf ("wrong tag found wanted %d found %d\n", tag, tlvTag);
				return (0);
			}
			return -1;	// not for us
		}
		if ( implicit != IMPLICIT_TAG ) // an explicit tag
		{
			L = decodeTL (P);	// get the explicit tag
			if ( L == -1 )
				return 0;
		}
	}	//P now points to the single TLV value 
	else
	{	// not a context tag
		if ( ( ( tlvTag & 0xc0 ) != ( tag & 0xc0 ) ) && Default )
			return -1;
	}
	return tlvTag;
}

/* functions to encode an oid */
/* X.Y.Z  etc
1st octet =X*40+Y
bit 7 =0 denotes last byte of sub identifier
2.100.3-> 81 34 03 hex
bit 8 is 0 for last 
2*40+100=180 dec= B4 hex= 10110100
padded out to two bytes   1000001 00110100 */
char *
decodeOID (OID oid) // goes from the binary form to a string
{
	char *buf, *tmp; // should be enough
	int i, charCount = 0;
	byte *p = oid.oidbytes;
	byte **pnt = &p;
	
	if (oid.length == 0)
		return NULL;
	
	i = getSubComponent (pnt);	// the first is 40* X +Y 
	buf = intToChar ( i / 40 );
	tmp = intToChar ( i % 40 );
	buf = ( char *)realloc( buf, strlen( buf ) + strlen( tmp ) + 2 );
	strcat( buf, "." );
	strcat( buf, tmp );
	free( tmp );

	while (p < (oid.oidbytes + oid.length))
	{
		int i = getSubComponent ( pnt );
		tmp = intToChar ( i );

		buf = ( char *)realloc( buf, strlen( buf ) + strlen( tmp ) + 2 );
		strcat( buf, "." );
		strcat( buf, tmp );
		free( tmp );
	}
	return buf;
}

int
getSubComponent (char **pnt)
// gets a sub component. If bit 8 is 0 it justgets the byte
// if bit8 is 1 then it keeps on going
{
	int result = 0;
	char *p = *pnt;

	while (*p & 0x80)
	{
		result = result * 128 + (*p++ & 0x7f);
	};

	result = result * 128 + (*p++ & 0x7f);
	*pnt = p;
	return result;
}

OID *
encodeOID (char *str)
//str is the 1.2.3.4 ascii string
{
	OID *oid;
	byte *result, *pnt;
	int resultindex = 0;
	int *buf, *temp;
	int x, length, len;
	int i, j = 0, sub = 0, noOfSubs;

	len = strlen( str );
	buf = ( int * )malloc( sizeof( int ) * (len +1 ) / 2 );
	temp = ( int * )malloc( sizeof( int ) * ( len +1 ) / 2 );
	result = ( byte * )malloc( sizeof( byte ) * ( len +1 ) / 2 );
	for (i = 0; i < len; i++)
	{
		if (str[i] != '.') {
			sub = sub * 10 + str[i] - '0';
		} else { // we have a .
			buf[j++] = sub;
			sub = 0;
		}
	}
	buf[j] = sub;
	noOfSubs = j + 1;
	// at this point buf contains each subcomponent as an integer
	// we now combine the first two
	buf[1] = buf[0] * 40 + buf[1];
	// we now just ignore buf[0]
	// now expand each int into the multi-byte form
	for (j = 1; j < noOfSubs; j++)
	{
		x = buf[j];
		for (i = 0; x > 0; i++) {
			temp[i] = x % 128;
			x /= 128;
		}
		length = i;
		for (i = 1; i < length; i++)
			temp[i] |= 0x80;
		for (i = 0; i < length; i++)
			result[resultindex++] = temp[length - i - 1];
	}
	oid = (OID *) malloc ( sizeof( OID ) );
	// malloc the storage for the encoded bytes
	pnt = (unsigned char *) malloc ( sizeof( unsigned char ) * ( resultindex + 1 ) );
	for (i = 0; i < resultindex; i++) {
		pnt[i] = result[i];
	}
	free( buf );
	free( temp );
	free( result );
	
	oid->length = resultindex;
	oid->oidbytes = pnt;
	return oid;
}

char *
xps_cat_names (char *name, char *type)
{
	char *s;
	int len;

	len = strlen (name) + strlen (type) + 2;
	s = (char *)malloc( sizeof( char ) * len );
	s = strcpy (s, name);
	s = strcat (s, ".");
	s = strcat (s, type);

	return s;
}

//malloc space for a string
char *
newString (char *str)
{
	char *pnt = (char *)malloc(sizeof( char ) * ( strlen (str) + 1 ) );
	strcpy (pnt, str);
	return pnt;
}

//TODO: comments. mallocs new space for b+a, adds them together & returns pointer
// concatenate RDNs with the ' , ' inbetween
char *
addString (char *a, char *b)
{
	int length;
	char *S;
	char *tmp;
	
	if ( b == NULL )
		return a;
		
	tmp = ( char *) malloc( strlen( a ) + 1 );
	tmp = strcpy ( tmp, a );
	length = strlen (a) + strlen (b) + 1;
	
	if ( a == NULL || *a =='\0' ) {
		S = (char *) realloc (a, length);
		S = strcpy ( S, b );
	} else {
		length += 1;
		S = (char *) realloc (a, length);
		S = strcpy ( S, b );
		S = strcat (S, ",");
	}
	S = strcat (S, tmp);
	free( tmp );
	return S;
}
/* TODO: escape charactars in DN according to  draft-ldapbis-dn */
char *escape_chars( char *dn )
{
	int i, j, len = strlen( dn ) + 1;
	int nlen = len;
	int dif = 'A' - 'a';
	char *ndn;
	for ( i = 0; i < len; i++ ) {
		if ( dn[i] ==' ' ||  dn[i] ==',' ||  dn[i] =='"' || dn[i] ==';' ||
			 dn[i] =='<' || dn[i] =='>' || dn[i] =='=' || dn[i] =='\\' )
			nlen++;
	}
	ndn = ( char * )malloc( sizeof( char * ) * nlen );
	for ( i = 0, j = 0; i < len; i++, j++ ) {
		if ( dn[i] ==' ' ||  dn[i] ==',' ||  dn[i] =='"' || dn[i] ==';' ||
			 dn[i] =='<' || dn[i] =='>' || dn[i] =='=' || dn[i] =='\\' )
		{
			ndn[j++] = '\\';
		}
		if ( ( dn[i] >='A' ) && ( dn[i] <='Z' ) )
		{
			ndn[j] = dn[i] - dif;
		} else
			ndn[j] = dn[i];
	}
	free( dn );
	return ndn;
}

char *
makeString (char *a, char *b, int length_b)
{				//mallocs a new string & copies a+b to it
	// e.g. S=makeString("c=",pnt,2); where pnt points to char array length 2
	int length, i, j;
	char *S, *tmp;
	S = strdup( a );
	tmp = ( char * )malloc( sizeof( char ) * ( length_b + 1 ) );
	for (i = 0; i < length_b; i++ )
		tmp[i] = b[i];
	tmp[length_b] = '\0';

	tmp = escape_chars( tmp );
	S = (char *)realloc( S, sizeof( char ) * ( strlen( S ) + strlen( tmp ) + 1 ) );
	strcat( S, tmp );
	free( tmp );
	return S;
}

// the generic routine to do c=  o= ou= etc
char *
r_d_n (char *addString, char *name, int tag, int implicit, byte ** P,
       int count, TypeDescription TD[], int optional, int Default,
       char *ObjectIdentifier )
{
	int i, length;
	byte *T;
	int tlvTag;
	long value;
	char *returnString;

	if (tag != -1) // we have a [tag]
	{
		if (implicit != IMPLICIT_TAG)
		{
			length = decodeTL (P);	// take out the outer tag
			if ( length == -1 )
				return NULL;
		}
	} else {	// tag == -1  we do not have a tag
		tlvTag = decodeTag (tag, implicit, P, optional, Default);
		if (tlvTag == -1)
			return newString ("");
	}

	length = decodeTL (P);
	if (length == -1 )
		return NULL;
	T = *P;
	returnString = makeString (addString, T, length);
	T += length;
	*P = T;
	return returnString;
}

char *
decodeLocality (char *name, int tag, int implicit, byte ** P, int count,
		TypeDescription TD[], int optional, int Default,
		char *ObjectIdentifier, Modifications *** xps_modtail, byte **P_CRL )
{
	char *S = r_d_n ("l=", name, tag, implicit, P, count, TD, optional,
			 Default, ObjectIdentifier);
	return S;
}

char *
decodeEmail (char *name, int tag, int implicit, byte ** P, int count,
	     TypeDescription TD[], int optional, int Default,
	     char *ObjectIdentifier, Modifications *** xps_modtail, byte **P_CRL )
{
	char *S =
		r_d_n ("email=", name, tag, implicit, P, count, TD, optional,
		       Default, ObjectIdentifier);
	return S;
}

char *
decodeCountry (char *name, int tag, int implicit, byte ** P, int count,
	       TypeDescription TD[], int optional, int Default,
	       char *ObjectIdentifier, Modifications *** xps_modtail, byte **P_CRL )
{
	char *S = r_d_n ("c=", name, tag, implicit, P, count, TD, optional,
			 Default, ObjectIdentifier);
	return S;
}

char *
decodeOrganisation (char *name, int tag, int implicit, byte ** P, int count,
		    TypeDescription TD[], int optional, int Default,
		    char *ObjectIdentifier, Modifications *** xps_modtail, byte **P_CRL )
{
	char *S = r_d_n ("o=", name, tag, implicit, P, count, TD, optional,
			 Default, ObjectIdentifier);
	return S;
}

char *
decodeOrganisationalUnit (char *name, int tag, int implicit, byte ** P,
			  int count, TypeDescription TD[], int optional,
			  int Default, char *ObjectIdentifier,
			  Modifications *** xps_modtail, byte **P_CRL )
{
	char *S = r_d_n ("ou=", name, tag, implicit, P, count, TD, optional,
			 Default, ObjectIdentifier);
	return S;
}

char *
decodeCommonName (char *name, int tag, int implicit, byte ** P, int count,
		  TypeDescription TD[], int optional, int Default,
		  char *ObjectIdentifier, Modifications *** xps_modtail, byte **P_CRL )
{
	char *S = r_d_n ("cn=", name, tag, implicit, P, count, TD, optional,
			 Default, ObjectIdentifier);
	return S;
}

char *
decodeDomainComponent (char *name, int tag, int implicit, byte ** P,
		       int count, TypeDescription TD[], int optional,
		       int Default, char *ObjectIdentifier,
		       Modifications *** xps_modtail, byte **P_CRL )
{
	char *S = r_d_n ("dc=", name, tag, implicit, P, count, TD, optional,
			 Default, ObjectIdentifier);
	return S;
}

char *
decodeStateOrProvince (char *name, int tag, int implicit, byte ** P, int count,
	       TypeDescription TD[], int optional, int Default,
	       char *ObjectIdentifier, Modifications *** xps_modtail, byte **P_CRL )
{
	char *S = r_d_n ("st=", name, tag, implicit, P, count, TD, optional,
			 Default, ObjectIdentifier);
	return S;
}

void
freeOID (OID * oid)
{
	free (oid->oidbytes);
	free (oid);
}

// replace the OIDREF anyDefinedBy by the latest value - freeing up the old value
void
addOID (char *Name, OID oid)
{
	lastOID.length = oid.length;
	lastOID.oidbytes = ( byte * )realloc( lastOID.oidbytes,
		sizeof( byte ) * oid.length );
	memcpy ( lastOID.oidbytes, oid.oidbytes, oid.length );
}

// returns 1 if they are equal 0 if they are different
int
oidCompare (OID a, OID b)
{
	return ( a.length == b.length &&
		!memcmp( a.oidbytes, b.oidbytes, a.length ) );
}

// oid contains the ber encoded Object Identifier
_CPFPLUS
getOidFunction()
{
	int i;
	OID *temp;

	// stepping through the table LookupTable[] defined above
	for (i = 0; LookupTable[i].function != NULL; i++ )
	{
		temp = encodeOID( LookupTable[i].oidString );
		if (oidCompare ( *temp, lastOID ) )
		{
			freeOID (temp);
			return LookupTable[i].function;
		}
		freeOID (temp);
	}
	return NULL;
}

char *
getOIDName()
{
	int i;
	OID *temp;

	// stepping through the table LookupTable[] defined above
	for (i = 0; LookupTable[i].function != NULL; i++ )
	{
		temp = encodeOID( LookupTable[i].oidString );
		if (oidCompare ( *temp, lastOID ) )
		{
			freeOID (temp);
			return LookupTable[i].name;
		}
		freeOID (temp);
	}
	return NULL;
}

char *
decodeEXTENSION (char *name, int tag, int implicit, byte ** P, int count,
		 TypeDescription TD[], int optional, int Default,
		 char *ObjectIdentifier, Modifications *** xps_modtail, byte **P_CRL )
{
	int i, j, length, tlvTag;
	long value;
	OIDREF *oidRef;
	byte *T, **ExtPointer;
	char *temp, *S;
	_CPFPLUS OF;

	tlvTag = decodeTag (tag, implicit, P, optional, Default);
	if (tlvTag == -1)
		return newString ("");
	
	length = decodeTL (P);
	if ( length == -1 )
		return NULL;
	
	T = *P;

	OF = getOidFunction ();
	if (OF == 0)
	{
		//printf ("\nExtension %s not defined in asn1 spec - skipping\n", z);
		T += length;
		*P = T;
		return newString ("");
	}

	ExtPointer = &T;
	temp = (char *) xps_cat_names( name, getOIDName () );

	S = OF (temp, tag, implicit, ExtPointer, length, 0, optional, Default,
	       getOIDName() /*ObjectIdentifier */ , xps_modtail, P_CRL );
	*P = T;
	free (S);
	free (temp);

	return newString ("");
}

char *
decodeSEQUENCE (char *name, int tag, int implicit, byte ** P, int count,
		TypeDescription TD[], int optional, int Default,
		char *ObjectIdentifer, Modifications *** xps_modtail, byte **P_CRL )
{
	int i, length, doneSoFar, tlvTag;
	byte *T, *start;
	char *returnString, *tempString;
	_CPFPLUS F;

	if (tag == -1)
		tag = 16;
	
	tlvTag = decodeTag (tag, implicit, P, optional, Default);
	if ( tlvTag == 0 ) 
		return NULL; // decodeTag failed
	if (tlvTag == -1)
		return newString( "" ); // SEQUENCE tag was expected, was not found, but this is not an error, if it is Optional
	
	length = decodeTL (P);
	if ( length == -1 ) {
		return NULL;
	}
	
	returnString = newString( "" );
	start = *P;
	for (i = 0; i < count; i++)
	{
		char *s = xps_cat_names (name, TD[i].name);

		F = (_CPFPLUS) TD[i].function;
		tempString =
			F (s, TD[i].tag, TD[i].tagOption, P, 0, 0,
			   TD[i].optional, TD[i].Default,
			   TD[i].ObjectIdentifier, xps_modtail, P_CRL );
		free (s);
		if (tempString == NULL )
		{	// we have an error in the called function
			free ( returnString );
			return NULL;
		}
		if (*tempString != '\0' ) // don't call addString if tempString is empty ("")
			returnString = addString (returnString, tempString); 
		
		free (tempString);

		doneSoFar = (int) (*P) - (int) start;
		if (doneSoFar == length)
		{
			return returnString;
		}
		if (doneSoFar > length)	// just a check for an over length
			break;
	}
	// reaching this point means an error
	free( returnString );
	return NULL;
}

char *
decodeSET (char *name, int tag, int implicit, byte ** P, int count,
	   TypeDescription TD[], int optional, int Default,
	   char *ObjectIdentifer, Modifications *** xps_modtail, byte **P_CRL )
{
	int i, length, doneSoFar;
	byte *T, *start;
	int tlvTag;
	_CPFPLUS F;
	char *returnString, *tempString;

	if (tag == -1)
		tag = 17;
	// get the SET tag
	tlvTag = decodeTag (tag, implicit, P, optional, Default);
	if ( tlvTag == -1 )
		return newString ("");
	
	length = decodeTL (P);
	if (length == 0 ||	length == -1 )
		return NULL;	// check for zero length set

	returnString = newString ("");
	start = *P;
	T = *P;
	tlvTag = *T; // get the tag
	do
	{
		// we must now step through the body comparing tags
		T = *P;
		tlvTag = *T;	// get the tag
		// compare it with the tags from the set
		for (i = 0; i < count; i++)
		{
			if (tlvTag == TD[i].tag)	// we have found the tag
			{
				F = (_CPFPLUS) TD[i].function;
				tempString =
					F (TD[i].name, TD[i].tag, TD[i].tagOption, P, 0, 0,
					   TD[i].optional, TD[i].Default,  TD[i].ObjectIdentifier,
					   xps_modtail, P_CRL );
				if ( tempString == NULL )
				{
					return NULL;
				}
				returnString = addString (returnString, tempString); //so far addString used only for creating DN - not required here
				free (tempString);
			}
		}
		doneSoFar = (int) (*P) - (int) start;
	} while (doneSoFar < length);

	return returnString;
}

char *
decodeCHOICE (char *name, int tag, int implicit, byte ** P, int count,
	      TypeDescription TD[], int optional, int Default,
	      char *ObjectIdentifier, Modifications *** xps_modtail, byte **P_CRL )
{
	int i, length;
	byte *T;
	int tlvTag;
	_CPFPLUS F;
	char *S;

	if (tag != -1)		// we have been called by a user tag- take it out
	{
		length = decodeTL (P);
		if ( length == -1 ) 
			return NULL;
	}
	T = *P;
	tlvTag = *T;

	// try and match the actual tlvTag with one of the functions
	for (i = 0; i < count; i++)
		if ((TD[i].tag) == tlvTag)
		{
			char *s;
			s = xps_cat_names (name, TD[0].name);
			F = (_CPFPLUS) TD[i].function;
			if (TD[i].tagOption != IMPLICIT_TAG)
			{
				decodeTL (P);	// if the tag is explicit remove it
				if ( length == -1 )
					return NULL;
			}
			S = F (s /*TD[i].name */ , TD[i].tag, TD[i].tagOption,
			       P, 0, 0, TD[i].optional, TD[i].Default,
			       TD[i].ObjectIdentifier, xps_modtail, P_CRL );
			free (s);
			return S;
		}

	// if the above does not work then the item may not be tagged
	// try and match the primitive instead
	for (i = 0; i < count; i++)
		if (TD[i].primitive == tlvTag)
		{
			char *s;
			s = (char *) xps_cat_names (name, TD[0].name);
			F = (_CPFPLUS) TD[i].function;
			S = F (s /*TD[i].name */ , TD[i].tag, TD[i].tagOption,
			       P, 0, 0, TD[i].optional, TD[i].Default,
			       TD[i].ObjectIdentifier, xps_modtail, P_CRL );
			free (s);
			return S;
		}

	if ( count == 1)		// try it anyway if there is only one
	{
		char *s;
		F = (_CPFPLUS) TD[0].function;
		s = (char *) xps_cat_names (name, TD[0].name);
		S = F (s, TD[0].tag, TD[0].tagOption, P, 0, 0,
		       TD[0].optional, TD[0].Default, TD[0].ObjectIdentifier,
		       xps_modtail, P_CRL );
		free (s);
		return S;
	}
	/* TODO: fixme - there is a bug in handling optional tags for CHOICE */
	if ( ( strcasecmp( name ,"certificateList.tbsCertList.nextUpdate") == 0 ) && optional ) {
		return newString (""); 
	}
		
	// giving up - cannot match a choice to the tag actually present
	return NULL;
		
}

char *
decodeSEQUENCE_OF (char *name, int tag, int implicit, byte ** P, int count,
		   TypeDescription TD[], int optional, int Default,
		   char *ObjectIdentifier, Modifications *** xps_modtail, byte **P_CRL )
{
	int i, length;
	byte *T;
	int tlvTag;
	int bytes_decoded = 0;
	byte *start;
	_CPFPLUS F;
	char *s;
	char *returnString, *tempString;

	
	if (tag == -1)
		tag = 16;
	tlvTag = decodeTag (tag, implicit, P, optional, Default);

	/* if we are about to call decode_5_1 ( decode sequence of revoked certificates)*/
	/* then save a pointer to the start of the sequence and skip it - it will be parsed later */
	if ( TD[0].function == decode_5_1 &&  *P_CRL == NULL) {
		*P_CRL = *P;
		length = decodeTL( P );
		*P+= length;
		return newString("");
	}

	length = decodeTL (P);
	if ( length == -1 )
		return NULL;
	
	start = *P;
	returnString = newString( "" );
	while ( ( i = ( ( int )*P - ( int )start ) ) < length )
	{
		F = (_CPFPLUS) TD[0].function;
		s = xps_cat_names (name, TD[0].name);
		tempString =
			F (s /*TD[0].name */ , TD[0].tag, TD[0].tagOption, P,
			   0, 0, TD[0].optional, TD[0].Default,
			   TD[0].ObjectIdentifier, xps_modtail, P_CRL );

		if ( tempString == NULL )	// an error
			return NULL;
		returnString = addString (returnString, tempString);
		free (tempString);
		free (s);
	}
	return returnString;
}

char *
decodeSET_OF (char *name, int tag, int implicit, byte ** P, int count,
	      TypeDescription TD[], int optional, int Default,
	      char *ObjectIdentifier, Modifications *** xps_modtail, byte **P_CRL )
{
	int i, length;
	byte *T;
	int tlvTag;
	int bytes_decoded = 0;
	byte *start;
	char *S, *returnString;
	_CPFPLUS F;

	if (tag == -1)
		tag = 49;
	tlvTag = decodeTag (tag, implicit, P, optional, Default);
	if ( tlvTag == -1 )
		return newString ("");
	if ( tlvTag == 0 )
		return NULL;

	length = decodeTL (P);
	if ( length == -1 )
		return NULL;
	start = *P;

	returnString = newString( "" );
	while ((i = ((int) *P - (int) start)) < length)
	{
		F = (_CPFPLUS) TD[0].function;
		S = F (name, TD[0].tag, TD[0].tagOption, P, 0, 0,
				TD[0].optional, TD[0].Default, TD[0].ObjectIdentifier,
				xps_modtail, P_CRL );
		if (S == 0)	// an error
			return 0;
		returnString = addString (returnString, S);
		free (S);
	}
	return returnString;
}

char *
decodeINTEGER (char *name, int tag, int implicit, byte ** P, int count,
	       TypeDescription TD[], int optional, int Default,
	       char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	int i, length, tlvTag;
	byte *T;
	long value, tmp_val;
	char *tmp;	
	ASN1_INTEGER *sn = NULL;
	if (tag == -1)
		tag = 2;
	
	tlvTag = decodeTag (tag, implicit, P, optional, Default);
	if ( tlvTag == -1 )
		return newString ("");
	if ( tlvTag == 0 )
		return NULL;
	
	length = decodeTL (P);
	if ( length == -1 )
		return NULL;
	
	T = *P;
	
	sn = ( ASN1_INTEGER * )malloc( sizeof( ASN1_INTEGER ) );
	
	sn->type = V_ASN1_INTEGER;
	sn->length = length;
	sn->data = T;
	sn->flags = 1;
	tmp = i2s_ASN1_INTEGER(0, sn );
	free( sn );	
	
	*P += length;

	xps_add_mod( name, tmp, strlen( tmp ), xps_modtail, 1 );

	free( tmp );
	return newString ("");
}

char *
decodeOBJECT_IDENTIFIER (char *name, int tag, int implicit, byte ** P,
			 int count, TypeDescription TD[], int optional,
			 int Default, char *ObjectIdentifier,
			 Modifications ***xps_modtail, byte **P_CRL )
{
	int length, tlvTag;
	OID oid;
	char *tmp;

	tlvTag = decodeTag ( tag, implicit, P, optional, Default );
	if ( tlvTag == -1 )
		return newString ( "" );
	if ( tlvTag == 0 )
		return NULL;

	length = decodeTL ( P );
	if ( length == -1 )
		return NULL;
	
	oid.length = length;
	oid.oidbytes = *P;
	addOID (name, oid);

	tmp = decodeOID (oid);
	xps_add_mod( name, tmp, strlen( tmp ), xps_modtail, 1 );
	free ( tmp);
	*P += length;
	return newString ("");
}

char *
decodeBOOLEAN (char *name, int tag, int implicit, byte ** P, int count,
	       TypeDescription TD[], int optional, int Default,
	       char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	int length, tlvTag;
	char *tmp;
	
	if (tag == -1)
		tag = 1;

	tlvTag = decodeTag ( tag, implicit, P, optional, Default );
	if ( tlvTag == -1 )
		return newString ( "" );
	if ( tlvTag == 0 )
		return NULL;

	length = decodeTL ( P );
	if ( length == -1 )
		return NULL;
	
	if ( ( **P ) && 1 )
		tmp = newString( "TRUE" );
	else
		tmp = newString( "FALSE" );
	
	xps_add_mod( name, tmp, strlen( tmp ), xps_modtail, 1 );

	free( tmp );
	*P += length;
	return newString ("");
}

char *
decodeANY_DEFINED_BY (char *name, int tag, int implicit, byte ** P, int count,
		      TypeDescription TD[], int optional, int Default,
		      char *ObjectIdentifier, Modifications *** xps_modtail, byte **P_CRL )
{
	_CPFPLUS OF;
	int j;

	// name is the dotted name built up during the call, but it has one too many components
	// e.g. Certificate.tbsCertificate.signature.parameters & we really just want the
	// Certificate.tbsCertificate.signature part in order to add the .algorithm part to it

	//TODO: remove or leave?	
	// find the end
	for (j = strlen (name) - 1; name[j] != '.'; j--);
		name[j] = 0;		// chop off the end
	OF = getOidFunction ();

	if (OF == NULL)
	{
		int length;
		/* decodeTLV was not called since there is nothing to be decoded and it will be skipped */
		length = decodeTL (P);
		if (length == -1 )
			return NULL;

		*P+=length;
		return newString ("");
	}
	return OF( name, tag, implicit, P, 1, 0, optional, Default,
		ObjectIdentifier, xps_modtail, P_CRL );
}

char *
decodeOCTET_STRING (char *name, int tag, int implicit, byte ** P, int count,
		    TypeDescription TD[], int optional, int Default,
		    char *ObjectIdentifier, Modifications *** xps_modtail, byte **P_CRL )
{
	int length, tlvTag;

	tlvTag = decodeTag ( tag, implicit, P, optional, Default );
	if ( tlvTag == -1 )
		return newString ( "" );
	if ( tlvTag == 0 )
		return NULL;

	length = decodeTL ( P );
	if ( length == -1 )
		return NULL;
	
	xps_add_mod( name, *P, length, xps_modtail, 0 );
	
	*P += length;
	return newString ("");
}

char *
decodePRINTABLESTRING (char *name, int tag, int implicit, byte **P,
		       int count, TypeDescription TD[], int optional,
		       int Default, char *ObjectIdentifier,
		       Modifications *** xps_modtail, byte **P_CRL )
{
	int length, tlvTag;
	char *str;

	tlvTag = decodeTag ( tag, implicit, P, optional, Default );
	if ( tlvTag == -1 )
		return newString ( "" );
	if ( tlvTag == 0 )
		return NULL;

	length = decodeTL ( P );
	if ( length == -1 )
		return NULL;
	str =( char *)strndup( *P, length );
	xps_add_mod( name, str, length, xps_modtail, 1 );
	free( str );
	*P += length;
	return newString ("");
}

char *
decodeUNIVERSALSTRING (char *name, int tag, int implicit, byte ** P,
		       int count, TypeDescription TD[], int optional,
		       int Default, char *ObjectIdentifier,
		       Modifications *** xps_modtail, byte **P_CRL )
{
	int length, tlvTag;
	char * str;
	
	tlvTag = decodeTag ( tag, implicit, P, optional, Default );
	if ( tlvTag == -1 )
		return newString ( "" );
	if ( tlvTag == 0 )
		return NULL;

	length = decodeTL ( P );
	if ( length == -1 )
		return NULL;
	str = ( char * )strndup( *P, length );
	xps_add_mod( name, str, length, xps_modtail, 1 );
	str = ( char * )strndup( *P, length );
	free( str );
	
	*P += length;
	return newString ("");
}

char *
decodeBMPSTRING (char *name, int tag, int implicit, byte ** P, int count,
		 TypeDescription TD[], int optional, int Default,
		 char *ObjectIdentifier, Modifications *** xps_modtail, byte **P_CRL )
{
	int length, tlvTag;
	char * str;

	tlvTag = decodeTag ( tag, implicit, P, optional, Default );
	if ( tlvTag == -1 )
		return newString ( "" );
	if ( tlvTag == 0 )
		return NULL;

	length = decodeTL ( P );
	if ( length == -1 )
		return NULL;
	str = ( char * )strndup( *P, length );
	xps_add_mod( name, str, length, xps_modtail, 1 );
	free( str );
	
	*P += length;
	return newString ("");
}

char *
decodeUTF8STRING (char *name, int tag, int implicit, byte ** P, int count,
		  TypeDescription TD[], int optional, int Default,
		  char *ObjectIdentifier, Modifications *** xps_modtail, byte **P_CRL )
{
	int length, tlvTag;

	tlvTag = decodeTag ( tag, implicit, P, optional, Default );
	if ( tlvTag == -1 )
		return newString ( "" );
	if ( tlvTag == 0 )
		return NULL;

	length = decodeTL ( P );
	if ( length == -1 )
		return NULL;
	
	xps_add_mod( name, *P, length, xps_modtail, 0 );
	
	*P += length;
	return newString ("");
}

char *
decodeIA5STRING (char *name, int tag, int implicit, byte ** P, int count,
		 TypeDescription TD[], int optional, int Default,
		 char *ObjectIdentifier, Modifications *** xps_modtail, byte **P_CRL )
{
	int length, tlvTag;
	char * str;

	tlvTag = decodeTag ( tag, implicit, P, optional, Default );
	if ( tlvTag == -1 )
		return newString ( "" );
	if ( tlvTag == 0 )
		return NULL;

	length = decodeTL ( P );
	if ( length == -1 )
		return NULL;
	
	str = ( char * )strndup( *P, length );
	xps_add_mod( name, str, length, xps_modtail, 1 );
	free( str );
	
	*P += length;
	return newString ("");
}

char *
decodeVISIBLESTRING (char *name, int tag, int implicit, byte ** P, int count,
		     TypeDescription TD[], int optional, int Default,
		     char *ObjectIdentifier, Modifications *** xps_modtail, byte **P_CRL )
{
	int length, tlvTag;
	char * str;

	tlvTag = decodeTag ( tag, implicit, P, optional, Default );
	if ( tlvTag == -1 )
		return newString ( "" );
	if ( tlvTag == 0 )
		return NULL;

	length = decodeTL ( P );
	if ( length == -1 )
		return NULL;
	
	str = ( char * )strndup( *P, length );
	xps_add_mod( name, str, length, xps_modtail, 1 );
	free( str );
	
	*P += length;
	return newString ("");
}

char *
decodeTELETEXSTRING (char *name, int tag, int implicit, byte ** P, int count,
		     TypeDescription TD[], int optional, int Default,
		     char *ObjectIdentifier, Modifications *** xps_modtail, byte **P_CRL )
{
	int length, tlvTag;
	char * str;

	tlvTag = decodeTag ( tag, implicit, P, optional, Default );
	if ( tlvTag == -1 )
		return newString ( "" );
	if ( tlvTag == 0 )
		return NULL;

	length = decodeTL ( P );
	if ( length == -1 )
		return NULL;
	
	str = ( char * )strndup( *P, length );
	xps_add_mod( name, str, length, xps_modtail, 1 );
	free( str );
	
	*P += length;
	return newString ("");
}

char *
decodeUTCTime (char *name, int tag, int implicit, byte ** P, int count,
	       TypeDescription TD[], int optional, int Default,
	       char *ObjectIdentifier, Modifications *** xps_modtail, byte **P_CRL )
{
	int length, tlvTag;
	char *buf;

	if (tag == -1)
		tag = 23;

	tlvTag = decodeTag ( tag, implicit, P, optional, Default );
	if ( tlvTag == -1 )
		return newString ( "" );
	if ( tlvTag == 0 )
		return NULL;

	length = decodeTL ( P );
	if ( length == -1 )
		return NULL;
	
	/* Since UTC time is not supported in the XPS, this value needs to be converted */
	/* to generalized time, i.e. we need 2 more characters for the year in the begining */
	buf = ( char * )malloc( sizeof (char) * length + 3);
	strncpy (buf + 2, (char *) *P, length);
	if (buf[2] > '4') {
		buf[0] = '1'; buf[1] = '9';
	} else {
		buf[0] = '2'; buf[1] = '0';
	}
	buf[length+2] = '\0';

	xps_add_mod( name, buf, length + 2, xps_modtail, 1 );
	*P += length;
	free ( buf );

	return newString ("");
}

char *
decodeGeneralizedTime (char *name, int tag, int implicit, byte **P,
		       int count, TypeDescription TD[], int optional,
		       int Default, char *ObjectIdentifier,
		       Modifications ***xps_modtail, byte **P_CRL )
{
	int length, tlvTag;
	char *str;

	if (tag == -1)
		tag = 24;

	tlvTag = decodeTag ( tag, implicit, P, optional, Default );
	if ( tlvTag == -1 )
		return newString ( "" );
	if ( tlvTag == 0 )
		return NULL;

	length = decodeTL ( P );
	if ( length == -1 )
		return NULL;
	
	str = ( char * )strndup( *P, length );
	xps_add_mod( name, str, length, xps_modtail, 1 );
	free( str );

	*P += length;
	return newString ("");
}

char *
decodeBIT_STRING (char *name, int tag, int implicit, byte ** P, int count,
		  TypeDescription TD[], int optional, int Default,
		  char *ObjectIdentifier, Modifications *** xps_modtail, byte **P_CRL )
{
	int length, tlvTag;

	if (tag == -1)		// there is no user tag
		tag = 3;

	tlvTag = decodeTag ( tag, implicit, P, optional, Default );
	if ( tlvTag == -1 )
		return newString ( "" );
	if ( tlvTag == 0 )
		return NULL;

	length = decodeTL ( P );
	if ( length == -1 )
		return NULL;
	xps_add_mod( name, *P, length, xps_modtail, 0 );
	
	*P += length;
	return newString ("");
}

char *
decodeENUMERATED (char *name, int tag, int implicit, byte ** P, int count,
		  TypeDescription TD[], int optional, int Default,
		  char *ObjectIdentifier, Modifications *** xps_modtail, byte **P_CRL )
{
	int i, length;
	byte *T;
	int tlvTag;
	long value;
	char *tmp;

	if (tag == -1)
		tag = 10;

	tlvTag = decodeTag ( tag, implicit, P, optional, Default );
	if ( tlvTag == -1 )
		return newString ( "" );
	if ( tlvTag == 0 )
		return NULL;

	length = decodeTL ( P );
	if ( length == -1 )
		return NULL;
	
	T = *P;
	for (i = 0, value = 0L; i < length; i++)
		value = value * 256 + *T++;
	*P = T;

	tmp = intToChar( value );
	xps_add_mod( name, tmp, strlen( tmp ), xps_modtail, 1 );

	free( tmp );
	return newString ("");
}

 // Definition C
char *decodeC(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeCountry(name,tag,tagOption,P,0,0,optional,Default,"2.5.4.6",xps_modtail, P_CRL );
	return S;
}
 // Definition O
char *decodeO(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeOrganisation(name,tag,tagOption,P,0,0,optional,Default,"2.5.4.10",xps_modtail, P_CRL );
	return S;
}
 // Definition Ou
char *decodeOu(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeOrganisationalUnit(name,tag,tagOption,P,0,0,optional,Default,"2.5.4.11",xps_modtail, P_CRL );
	return S;
}
 // Definition Cn
char *decodeCn(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeCommonName(name,tag,tagOption,P,0,0,optional,Default,"2.5.4.3",xps_modtail, P_CRL );
	return S;
}
 // Definition L
char *decodeL(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeLocality(name,tag,tagOption,P,0,0,optional,Default,"2.5.4.7",xps_modtail, P_CRL );
	return S;
}
 // Definition St
char *decodeSt(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeStateOrProvince(name,tag,tagOption,P,0,0,optional,Default,"2.5.4.8",xps_modtail, P_CRL );
	return S;
}
 // Definition E_mail
char *decodeE_mail(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeEmail(name,tag,tagOption,P,0,0,optional,Default,"2.5.4.99",xps_modtail, P_CRL );
	return S;
}
 // Definition Dc
char *decodeDc(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeDomainComponent(name,tag,tagOption,P,0,0,optional,Default,"0.9.2342.19200300.100.1.25",xps_modtail, P_CRL );
	return S;
}
 // Definition AttributeCertificate
char *decodeAttributeCertificate(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"acinfo",-1,0,0,(_CPFPLUS)decodeAttributeCertificateInfo,0,0,""};
	TypeDescription _TD1 = {"signatureAlgorithm",-1,0,0,(_CPFPLUS)decodeAlgorithmIdentifier,0,0,""};
	TypeDescription _TD2 = {"signatureValue",-1,0,0,(_CPFPLUS)decodeBIT_STRING,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2};

	S = decodeSEQUENCE(name,tag,tagOption,P,3,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition AttributeCertificateInfo
char *decodeAttributeCertificateInfo(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"version",-1,0,0,(_CPFPLUS)decodeAttCertVersion,0,1,""};
	TypeDescription _TD1 = {"holder",-1,0,0,(_CPFPLUS)decodeHolder,0,0,""};
	TypeDescription _TD2 = {"issuer",-1,0,0,(_CPFPLUS)decodeAttCertIssuer,0,0,""};
	TypeDescription _TD3 = {"signature",-1,0,0,(_CPFPLUS)decodeAlgorithmIdentifier,0,0,""};
	TypeDescription _TD4 = {"serialNumber",-1,0,0,(_CPFPLUS)decodeCertificateSerialNumber,0,0,""};
	TypeDescription _TD5 = {"attrCertValidityPeriod",-1,0,0,(_CPFPLUS)decodeAttCertValidityPeriod,0,0,""};
	TypeDescription _TD6 = {"attributes",-1,0,0,(_CPFPLUS)decodeAttributes,0,0,""};
	TypeDescription _TD7 = {"issuerUniqueID",-1,0,0,(_CPFPLUS)decodeUniqueIdentifier,1,0,""};
	TypeDescription _TD8 = {"extensions",-1,0,0,(_CPFPLUS)decodeExtensions,1,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2,_TD3,_TD4,_TD5,_TD6,_TD7,_TD8};

	S = decodeSEQUENCE(name,tag,tagOption,P,9,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition _0
char *decode_0(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"",-1,0,0,(_CPFPLUS)decodeAttribute,0,0,""};
	TypeDescription _TD[] = {_TD0};

	S = decodeSEQUENCE_OF(name,tag,tagOption,P,1,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition Attributes
char *decodeAttributes(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decode_0(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition AttCertVersion
char *decodeAttCertVersion(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeINTEGER(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition Holder
char *decodeHolder(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"baseCertificateID",160,0,1,(_CPFPLUS)decodeIssuerSerial,1,0,""};
	TypeDescription _TD1 = {"entityName",161,0,1,(_CPFPLUS)decodeGeneralNames,1,0,""};
	TypeDescription _TD2 = {"objectDigestInfo",162,0,1,(_CPFPLUS)decodeObjectDigestInfo,1,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2};

	S = decodeSEQUENCE(name,tag,tagOption,P,3,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition ObjectDigestInfo
char *decodeObjectDigestInfo(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"digestedObjectType",-1,10,0,(_CPFPLUS)decodeENUMERATED,0,0,""};
	TypeDescription _TD1 = {"otherObjectTypeID",-1,0,0,(_CPFPLUS)decodeOBJECT_IDENTIFIER,1,0,""};
	TypeDescription _TD2 = {"digestAlgorithm",-1,0,0,(_CPFPLUS)decodeAlgorithmIdentifier,0,0,""};
	TypeDescription _TD3 = {"objectDigest",-1,0,0,(_CPFPLUS)decodeBIT_STRING,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2,_TD3};

	S = decodeSEQUENCE(name,tag,tagOption,P,4,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition AttCertIssuer
char *decodeAttCertIssuer(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"v2Form",160,0,1,(_CPFPLUS)decodeV2Form,0,0,""};
	TypeDescription _TD[] = {_TD0};

	S = decodeCHOICE(name,tag,tagOption,P,1,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition V2Form
char *decodeV2Form(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"issuerName",-1,0,0,(_CPFPLUS)decodeGeneralNames,0,0,""};
	TypeDescription _TD1 = {"baseCertificateID",160,0,1,(_CPFPLUS)decodeIssuerSerial,1,0,""};
	TypeDescription _TD2 = {"objectDigestInfo",161,0,1,(_CPFPLUS)decodeObjectDigestInfo,1,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2};

	S = decodeSEQUENCE(name,tag,tagOption,P,3,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition IssuerSerial
char *decodeIssuerSerial(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"issuer",-1,0,0,(_CPFPLUS)decodeGeneralNames,0,0,""};
	TypeDescription _TD1 = {"serial",-1,0,0,(_CPFPLUS)decodeCertificateSerialNumber,0,0,""};
	TypeDescription _TD2 = {"issuerUID",-1,0,0,(_CPFPLUS)decodeUniqueIdentifier,1,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2};

	S = decodeSEQUENCE(name,tag,tagOption,P,3,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition AttCertValidityPeriod
char *decodeAttCertValidityPeriod(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"notBeforeTime",-1,24,0,(_CPFPLUS)decodeGeneralizedTime,0,0,""};
	TypeDescription _TD1 = {"notAfterTime",-1,24,0,(_CPFPLUS)decodeGeneralizedTime,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1};

	S = decodeSEQUENCE(name,tag,tagOption,P,2,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition AttributeTypeAndValue
char *decodeAttributeTypeAndValue(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"type",-1,0,0,(_CPFPLUS)decodeOBJECT_IDENTIFIER,0,0,""};
	TypeDescription _TD1 = {"value",-1,0,0,(_CPFPLUS)decodeANY_DEFINED_BY,0,0,"type"};
	TypeDescription _TD[] = {_TD0,_TD1};

	S = decodeSEQUENCE(name,tag,tagOption,P,2,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition _1
char *decode_1(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"",-1,0,0,(_CPFPLUS)decodeANY_DEFINED_BY,0,0,"type"};
	TypeDescription _TD[] = {_TD0};

	S = decodeSET_OF(name,tag,tagOption,P,1,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition Attribute
char *decodeAttribute(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"type",-1,0,0,(_CPFPLUS)decodeOBJECT_IDENTIFIER,0,0,""};
	TypeDescription _TD1 = {"values",-1,0,0,(_CPFPLUS)decode_1,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1};

	S = decodeSEQUENCE(name,tag,tagOption,P,2,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition X520name
char *decodeX520name(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"teletexString",-1,0,0,(_CPFPLUS)decodeTELETEXSTRING,0,0,""};
	TypeDescription _TD1 = {"printableString",-1,0,0,(_CPFPLUS)decodePRINTABLESTRING,0,0,""};
	TypeDescription _TD2 = {"universalString",-1,0,0,(_CPFPLUS)decodeUNIVERSALSTRING,0,0,""};
	TypeDescription _TD3 = {"utf8String",-1,0,0,(_CPFPLUS)decodeUTF8STRING,0,0,""};
	TypeDescription _TD4 = {"bmpString",-1,0,0,(_CPFPLUS)decodeBMPSTRING,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2,_TD3,_TD4};

	S = decodeCHOICE(name,tag,tagOption,P,5,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition X520CommonName
char *decodeX520CommonName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"teletexString",-1,0,0,(_CPFPLUS)decodeTELETEXSTRING,0,0,""};
	TypeDescription _TD1 = {"printableString",-1,0,0,(_CPFPLUS)decodePRINTABLESTRING,0,0,""};
	TypeDescription _TD2 = {"universalString",-1,0,0,(_CPFPLUS)decodeUNIVERSALSTRING,0,0,""};
	TypeDescription _TD3 = {"utf8String",-1,0,0,(_CPFPLUS)decodeUTF8STRING,0,0,""};
	TypeDescription _TD4 = {"bmpString",-1,0,0,(_CPFPLUS)decodeBMPSTRING,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2,_TD3,_TD4};

	S = decodeCHOICE(name,tag,tagOption,P,5,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition X520LocalityName
char *decodeX520LocalityName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"teletexString",-1,0,0,(_CPFPLUS)decodeTELETEXSTRING,0,0,""};
	TypeDescription _TD1 = {"printableString",-1,0,0,(_CPFPLUS)decodePRINTABLESTRING,0,0,""};
	TypeDescription _TD2 = {"universalString",-1,0,0,(_CPFPLUS)decodeUNIVERSALSTRING,0,0,""};
	TypeDescription _TD3 = {"utf8String",-1,0,0,(_CPFPLUS)decodeUTF8STRING,0,0,""};
	TypeDescription _TD4 = {"bmpString",-1,0,0,(_CPFPLUS)decodeBMPSTRING,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2,_TD3,_TD4};

	S = decodeCHOICE(name,tag,tagOption,P,5,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition X520StateOrProvinceName
char *decodeX520StateOrProvinceName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"teletexString",-1,0,0,(_CPFPLUS)decodeTELETEXSTRING,0,0,""};
	TypeDescription _TD1 = {"printableString",-1,0,0,(_CPFPLUS)decodePRINTABLESTRING,0,0,""};
	TypeDescription _TD2 = {"universalString",-1,0,0,(_CPFPLUS)decodeUNIVERSALSTRING,0,0,""};
	TypeDescription _TD3 = {"utf8String",-1,0,0,(_CPFPLUS)decodeUTF8STRING,0,0,""};
	TypeDescription _TD4 = {"bmpString",-1,0,0,(_CPFPLUS)decodeBMPSTRING,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2,_TD3,_TD4};

	S = decodeCHOICE(name,tag,tagOption,P,5,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition X520OrganizationName
char *decodeX520OrganizationName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"teletexString",-1,0,0,(_CPFPLUS)decodeTELETEXSTRING,0,0,""};
	TypeDescription _TD1 = {"printableString",-1,0,0,(_CPFPLUS)decodePRINTABLESTRING,0,0,""};
	TypeDescription _TD2 = {"universalString",-1,0,0,(_CPFPLUS)decodeUNIVERSALSTRING,0,0,""};
	TypeDescription _TD3 = {"utf8String",-1,0,0,(_CPFPLUS)decodeUTF8STRING,0,0,""};
	TypeDescription _TD4 = {"bmpString",-1,0,0,(_CPFPLUS)decodeBMPSTRING,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2,_TD3,_TD4};

	S = decodeCHOICE(name,tag,tagOption,P,5,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition X520OrganizationalUnitName
char *decodeX520OrganizationalUnitName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"teletexString",-1,0,0,(_CPFPLUS)decodeTELETEXSTRING,0,0,""};
	TypeDescription _TD1 = {"printableString",-1,0,0,(_CPFPLUS)decodePRINTABLESTRING,0,0,""};
	TypeDescription _TD2 = {"universalString",-1,0,0,(_CPFPLUS)decodeUNIVERSALSTRING,0,0,""};
	TypeDescription _TD3 = {"utf8String",-1,0,0,(_CPFPLUS)decodeUTF8STRING,0,0,""};
	TypeDescription _TD4 = {"bmpString",-1,0,0,(_CPFPLUS)decodeBMPSTRING,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2,_TD3,_TD4};

	S = decodeCHOICE(name,tag,tagOption,P,5,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition X520Title
char *decodeX520Title(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"teletexString",-1,0,0,(_CPFPLUS)decodeTELETEXSTRING,0,0,""};
	TypeDescription _TD1 = {"printableString",-1,0,0,(_CPFPLUS)decodePRINTABLESTRING,0,0,""};
	TypeDescription _TD2 = {"universalString",-1,0,0,(_CPFPLUS)decodeUNIVERSALSTRING,0,0,""};
	TypeDescription _TD3 = {"utf8String",-1,0,0,(_CPFPLUS)decodeUTF8STRING,0,0,""};
	TypeDescription _TD4 = {"bmpString",-1,0,0,(_CPFPLUS)decodeBMPSTRING,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2,_TD3,_TD4};

	S = decodeCHOICE(name,tag,tagOption,P,5,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition X520dnQualifier
char *decodeX520dnQualifier(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodePRINTABLESTRING(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition X520countryName
char *decodeX520countryName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodePRINTABLESTRING(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition X520SerialNumber
char *decodeX520SerialNumber(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodePRINTABLESTRING(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition X520Pseudonym
char *decodeX520Pseudonym(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"teletexString",-1,0,0,(_CPFPLUS)decodeTELETEXSTRING,0,0,""};
	TypeDescription _TD1 = {"printableString",-1,0,0,(_CPFPLUS)decodePRINTABLESTRING,0,0,""};
	TypeDescription _TD2 = {"universalString",-1,0,0,(_CPFPLUS)decodeUNIVERSALSTRING,0,0,""};
	TypeDescription _TD3 = {"utf8String",-1,0,0,(_CPFPLUS)decodeUTF8STRING,0,0,""};
	TypeDescription _TD4 = {"bmpString",-1,0,0,(_CPFPLUS)decodeBMPSTRING,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2,_TD3,_TD4};

	S = decodeCHOICE(name,tag,tagOption,P,5,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition X520DomainComponent
char *decodeX520DomainComponent(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeIA5STRING(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition EmailAddress
char *decodeEmailAddress(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeIA5STRING(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition Name
char *decodeName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"rdnSequence",-1,0,0,(_CPFPLUS)decodeRDNSequence,0,0,""};
	TypeDescription _TD[] = {_TD0};

	S = decodeCHOICE(name,tag,tagOption,P,1,_TD,optional,Default,"", xps_modtail, P_CRL );
//	S = escape_chars( S );
	xps_add_mod( name, S, strlen( S ), xps_modtail, 1 );

	return S;
}
 // Definition _2
char *decode_2(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"",-1,0,0,(_CPFPLUS)decodeRelativeDistinguishedName,0,0,""};
	TypeDescription _TD[] = {_TD0};

	S = decodeSEQUENCE_OF(name,tag,tagOption,P,1,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition RDNSequence
char *decodeRDNSequence(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decode_2(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition DistinguishedName
char *decodeDistinguishedName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeRDNSequence(name,tag,tagOption,P,0,0,optional,Default,"", xps_modtail, P_CRL );

	return S;
}
 // Definition _3
char *decode_3(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"",-1,0,0,(_CPFPLUS)decodeAttributeTypeAndValue,0,0,""};
	TypeDescription _TD[] = {_TD0};

	S = decodeSET_OF(name,tag,tagOption,P,1,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition RelativeDistinguishedName
char *decodeRelativeDistinguishedName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decode_3(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition DirectoryString
char *decodeDirectoryString(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"teletexString",-1,0,0,(_CPFPLUS)decodeTELETEXSTRING,0,0,""};
	TypeDescription _TD1 = {"printableString",-1,0,0,(_CPFPLUS)decodePRINTABLESTRING,0,0,""};
	TypeDescription _TD2 = {"universalString",-1,0,0,(_CPFPLUS)decodeUNIVERSALSTRING,0,0,""};
	TypeDescription _TD3 = {"utf8String",-1,0,0,(_CPFPLUS)decodeUTF8STRING,0,0,""};
	TypeDescription _TD4 = {"bmpString",-1,0,0,(_CPFPLUS)decodeBMPSTRING,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2,_TD3,_TD4};

	S = decodeCHOICE(name,tag,tagOption,P,5,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition Certificate
char *decodeCertificate(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"tbsCertificate",-1,0,0,(_CPFPLUS)decodeTBSCertificate,0,0,""};
	TypeDescription _TD1 = {"signatureAlgorithm",-1,0,0,(_CPFPLUS)decodeAlgorithmIdentifier,0,0,""};
	TypeDescription _TD2 = {"signature",-1,0,0,(_CPFPLUS)decodeBIT_STRING,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2};

	S = decodeSEQUENCE(name,tag,tagOption,P,3,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition TBSCertificate
char *decodeTBSCertificate(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"version",160,0,0,(_CPFPLUS)decodeVersion,0,1,""};
	TypeDescription _TD1 = {"serialNumber",-1,0,0,(_CPFPLUS)decodeCertificateSerialNumber,0,0,""};
	TypeDescription _TD2 = {"signature",-1,0,0,(_CPFPLUS)decodeAlgorithmIdentifier,0,0,""};
	TypeDescription _TD3 = {"issuer",-1,0,0,(_CPFPLUS)decodeName,0,0,""};
	TypeDescription _TD4 = {"validity",-1,0,0,(_CPFPLUS)decodeValidity,0,0,""};
	TypeDescription _TD5 = {"subject",-1,0,0,(_CPFPLUS)decodeName,0,0,""};
	TypeDescription _TD6 = {"subjectPublicKeyInfo",-1,0,0,(_CPFPLUS)decodeSubjectPublicKeyInfo,0,0,""};
	TypeDescription _TD7 = {"issuerUniqueID",161,0,1,(_CPFPLUS)decodeUniqueIdentifier,1,0,""};
	TypeDescription _TD8 = {"subjectUniqueID",162,0,1,(_CPFPLUS)decodeUniqueIdentifier,1,0,""};
	TypeDescription _TD9 = {"extensions",163,0,0,(_CPFPLUS)decodeExtensions,1,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2,_TD3,_TD4,_TD5,_TD6,_TD7,_TD8,_TD9};

	S = decodeSEQUENCE(name,tag,tagOption,P,10,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition Version
char *decodeVersion(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeINTEGER(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition CertificateSerialNumber
char *decodeCertificateSerialNumber(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeINTEGER(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition Validity
char *decodeValidity(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"notBefore",-1,0,0,(_CPFPLUS)decodeTime,0,0,""};
	TypeDescription _TD1 = {"notAfter",-1,0,0,(_CPFPLUS)decodeTime,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1};

	S = decodeSEQUENCE(name,tag,tagOption,P,2,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition Time
char *decodeTime(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"utcTime",-1,23,0,(_CPFPLUS)decodeUTCTime,0,0,""};
	TypeDescription _TD1 = {"generalTime",-1,24,0,(_CPFPLUS)decodeGeneralizedTime,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1};

	S = decodeCHOICE(name,tag,tagOption,P,2,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition UniqueIdentifier
char *decodeUniqueIdentifier(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeBIT_STRING(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition SubjectPublicKeyInfo
char *decodeSubjectPublicKeyInfo(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"algorithm",-1,0,0,(_CPFPLUS)decodeAlgorithmIdentifier,0,0,""};
	TypeDescription _TD1 = {"subjectPublicKey",-1,0,0,(_CPFPLUS)decodeBIT_STRING,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1};

	S = decodeSEQUENCE(name,tag,tagOption,P,2,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition _4
char *decode_4(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"",-1,0,0,(_CPFPLUS)decodeExtension,0,0,""};
	TypeDescription _TD[] = {_TD0};

	S = decodeSEQUENCE_OF(name,tag,tagOption,P,1,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition Extensions
char *decodeExtensions(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decode_4(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition Extension
char *decodeExtension(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"extnID",-1,0,0,(_CPFPLUS)decodeOBJECT_IDENTIFIER,0,0,""};
	TypeDescription _TD1 = {"critical",-1,1,0,(_CPFPLUS)decodeBOOLEAN,0,1,""};
	TypeDescription _TD2 = {"extnValue",-1,0,0,(_CPFPLUS)decodeEXTENSION,0,0,"extnID"};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2};

	S = decodeSEQUENCE(name,tag,tagOption,P,3,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition CertificateList
char *decodeCertificateList(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"tbsCertList",-1,0,0,(_CPFPLUS)decodeTBSCertList,0,0,""};
	TypeDescription _TD1 = {"signatureAlgorithm",-1,0,0,(_CPFPLUS)decodeAlgorithmIdentifier,0,0,""};
	TypeDescription _TD2 = {"signature",-1,0,0,(_CPFPLUS)decodeBIT_STRING,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2};

	S = decodeSEQUENCE(name,tag,tagOption,P,3,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}

char *decode_5_1(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"userCertificate",-1,0,0,(_CPFPLUS)decodeCertificateSerialNumber,0,0,""};
	TypeDescription _TD1 = {"revocationDate",-1,0,0,(_CPFPLUS)decodeTime,0,0,""};
	TypeDescription _TD2 = {"crlEntryExtensions",-1,0,0,(_CPFPLUS)decodeExtensions,1,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2};
	// decode SEQUENCE OF Revoked Cerificates
	//rs->sr_err = xps_x509attrs2entries( op, rs, mod,
	//	oc_type, abs_oc_type, dnflag, dn,
	//	decodeSEQUENCE_OF, decodeName, e );
	
	S = decodeSEQUENCE(name,tag,tagOption,P,3,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
// Definition _5
char *decode_5(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"",-1,16,0,(_CPFPLUS)decode_5_1,0,0,""};
	TypeDescription _TD[] = {_TD0};
	// decode SEQUENCE OF Revoked Cerificates
	S = decodeSEQUENCE_OF(name,tag,tagOption,P,1,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition TBSCertList
char *decodeTBSCertList(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"version",-1,0,0,(_CPFPLUS)decodeVersion,1,0,""};
	TypeDescription _TD1 = {"signature",-1,0,0,(_CPFPLUS)decodeAlgorithmIdentifier,0,0,""};
	TypeDescription _TD2 = {"issuer",-1,0,0,(_CPFPLUS)decodeName,0,0,""};
	TypeDescription _TD3 = {"thisUpdate",-1,0,0,(_CPFPLUS)decodeTime,0,0,""};
	TypeDescription _TD4 = {"nextUpdate",-1,0,0,(_CPFPLUS)decodeTime,1,0,""};
	TypeDescription _TD5 = {"revokedCertificates",-1,0,0,(_CPFPLUS)decode_5,1,0,""};
	TypeDescription _TD6 = {"crlExtensions",160,0,0,(_CPFPLUS)decodeExtensions,1,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2,_TD3,_TD4,_TD5,_TD6};

	S = decodeSEQUENCE(name,tag,tagOption,P,7,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition AlgorithmIdentifier
char *decodeAlgorithmIdentifier(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"algorithm",-1,0,0,(_CPFPLUS)decodeOBJECT_IDENTIFIER,0,0,""};
	TypeDescription _TD1 = {"parameters",-1,0,0,(_CPFPLUS)decodeANY_DEFINED_BY,1,0,"algorithm"};
	TypeDescription _TD[] = {_TD0,_TD1};

	S = decodeSEQUENCE(name,tag,tagOption,P,2,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition AuthorityKeyIdentifier
char *decodeAuthorityKeyIdentifier(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"keyIdentifier",160,0,1,(_CPFPLUS)decodeKeyIdentifier,1,0,""};
	TypeDescription _TD1 = {"authorityCertIssuer",161,0,1,(_CPFPLUS)decodeGeneralNames,1,0,""};
	TypeDescription _TD2 = {"authorityCertSerialNumber",162,0,1,(_CPFPLUS)decodeCertificateSerialNumber,1,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2};

	S = decodeSEQUENCE(name,tag,tagOption,P,3,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition KeyIdentifier
char *decodeKeyIdentifier(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeOCTET_STRING(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition SubjectKeyIdentifier
char *decodeSubjectKeyIdentifier(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeKeyIdentifier(name,tag,tagOption,P,0,0,optional,Default,"2.5.29.14",xps_modtail, P_CRL );
	return S;
}
 // Definition KeyUsage
char *decodeKeyUsage(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeBIT_STRING(name,tag,tagOption,P,0,0,optional,Default,"2.5.29.15",xps_modtail, P_CRL );
	return S;
}
 // Definition PrivateKeyUsagePeriod
char *decodePrivateKeyUsagePeriod(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"notBefore",128,24,1,(_CPFPLUS)decodeGeneralizedTime,1,0,""};
	TypeDescription _TD1 = {"notAfter",129,24,1,(_CPFPLUS)decodeGeneralizedTime,1,0,""};
	TypeDescription _TD[] = {_TD0,_TD1};

	S = decodeSEQUENCE(name,tag,tagOption,P,2,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition _6
char *decode_6(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"",-1,0,0,(_CPFPLUS)decodePolicyInformation,0,0,""};
	TypeDescription _TD[] = {_TD0};

	S = decodeSEQUENCE_OF(name,tag,tagOption,P,1,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition CertificatePolicies
char *decodeCertificatePolicies(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decode_6(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition _7
char *decode_7(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"",-1,0,0,(_CPFPLUS)decodePolicyQualifierInfo,0,0,""};
	TypeDescription _TD[] = {_TD0};

	S = decodeSEQUENCE_OF(name,tag,tagOption,P,1,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition PolicyInformation
char *decodePolicyInformation(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"policyIdentifier",-1,0,0,(_CPFPLUS)decodeCertPolicyId,0,0,""};
	TypeDescription _TD1 = {"policyQualifiers",-1,0,0,(_CPFPLUS)decode_7,1,0,""};
	TypeDescription _TD[] = {_TD0,_TD1};

	S = decodeSEQUENCE(name,tag,tagOption,P,2,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition CertPolicyId
char *decodeCertPolicyId(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeOBJECT_IDENTIFIER(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition PolicyQualifierInfo
char *decodePolicyQualifierInfo(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"policyQualifierId",-1,0,0,(_CPFPLUS)decodeOBJECT_IDENTIFIER,0,0,""};
	TypeDescription _TD1 = {"qualifier",-1,0,0,(_CPFPLUS)decodeANY_DEFINED_BY,0,0,"policyQualifierId"};
	TypeDescription _TD[] = {_TD0,_TD1};

	S = decodeSEQUENCE(name,tag,tagOption,P,2,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition UserNotice
char *decodeUserNotice(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"noticeRef",-1,0,0,(_CPFPLUS)decodeNoticeReference,1,0,""};
	TypeDescription _TD1 = {"explicitText",-1,0,0,(_CPFPLUS)decodeDisplayText,1,0,""};
	TypeDescription _TD[] = {_TD0,_TD1};

	S = decodeSEQUENCE(name,tag,tagOption,P,2,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition _8
char *decode_8(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"",-1,2,0,(_CPFPLUS)decodeINTEGER,0,0,""};
	TypeDescription _TD[] = {_TD0};

	S = decodeSEQUENCE_OF(name,tag,tagOption,P,1,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition NoticeReference
char *decodeNoticeReference(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"organization",-1,0,0,(_CPFPLUS)decodeDisplayText,0,0,""};
	TypeDescription _TD1 = {"noticeNumbers",-1,0,0,(_CPFPLUS)decode_8,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1};

	S = decodeSEQUENCE(name,tag,tagOption,P,2,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition DisplayText
char *decodeDisplayText(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"ia5String",-1,0,0,(_CPFPLUS)decodeIA5STRING,0,0,""};
	TypeDescription _TD1 = {"visibleString",-1,0,0,(_CPFPLUS)decodeVISIBLESTRING,0,0,""};
	TypeDescription _TD2 = {"bmpString",-1,0,0,(_CPFPLUS)decodeBMPSTRING,0,0,""};
	TypeDescription _TD3 = {"utf8String",-1,0,0,(_CPFPLUS)decodeUTF8STRING,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2,_TD3};

	S = decodeCHOICE(name,tag,tagOption,P,4,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition _9
char *decode_9(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"",-1,16,0,(_CPFPLUS)decodeSEQUENCE,0,0,""};
	TypeDescription _TD1 = {"issuerDomainPolicy",-1,0,0,(_CPFPLUS)decodeCertPolicyId,0,0,""};
	TypeDescription _TD2 = {"subjectDomainPolicy",-1,0,0,(_CPFPLUS)decodeCertPolicyId,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2};

	S = decodeSEQUENCE_OF(name,tag,tagOption,P,3,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition PolicyMappings
char *decodePolicyMappings(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decode_9(name,tag,tagOption,P,0,0,optional,Default,"2.5.29.33",xps_modtail, P_CRL );
	return S;
}
 // Definition SubjectAltName
char *decodeSubjectAltName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeGeneralNames(name,tag,tagOption,P,0,0,optional,Default,"2.5.29.17",xps_modtail, P_CRL );
	return S;
}
 // Definition _10
char *decode_10(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"",-1,0,0,(_CPFPLUS)decodeGeneralName,0,0,""};
	TypeDescription _TD[] = {_TD0};

	S = decodeSEQUENCE_OF(name,tag,tagOption,P,1,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition GeneralNames
char *decodeGeneralNames(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decode_10(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition GeneralName
char *decodeGeneralName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"otherName",160,0,1,(_CPFPLUS)decodeAnotherName,0,0,""};
	TypeDescription _TD1 = {"rfc822Name",129,0,1,(_CPFPLUS)decodeIA5STRING,0,0,""};
	TypeDescription _TD2 = {"dNSName",130,0,1,(_CPFPLUS)decodeIA5STRING,0,0,""};
	TypeDescription _TD3 = {"directoryName",164,0,1,(_CPFPLUS)decodeName,0,0,""};
	TypeDescription _TD4 = {"ediPartyName",165,0,1,(_CPFPLUS)decodeEDIPartyName,0,0,""};
	TypeDescription _TD5 = {"uniformResourceIdentifier",134,0,1,(_CPFPLUS)decodeIA5STRING,0,0,""};
	TypeDescription _TD6 = {"iPAddress",135,0,1,(_CPFPLUS)decodeOCTET_STRING,0,0,""};
	TypeDescription _TD7 = {"registeredID",168,0,1,(_CPFPLUS)decodeOBJECT_IDENTIFIER,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2,_TD3,_TD4,_TD5,_TD6,_TD7};

	S = decodeCHOICE(name,tag,tagOption,P,8,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition AnotherName
char *decodeAnotherName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"type-id",-1,0,0,(_CPFPLUS)decodeOBJECT_IDENTIFIER,0,0,""};
	TypeDescription _TD1 = {"value",160,0,0,(_CPFPLUS)decodeANY_DEFINED_BY,0,0,"type-id"};
	TypeDescription _TD[] = {_TD0,_TD1};

	S = decodeSEQUENCE(name,tag,tagOption,P,2,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition EDIPartyName
char *decodeEDIPartyName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"nameAssigner",160,0,1,(_CPFPLUS)decodeDirectoryString,1,0,""};
	TypeDescription _TD1 = {"partyName",161,0,1,(_CPFPLUS)decodeDirectoryString,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1};

	S = decodeSEQUENCE(name,tag,tagOption,P,2,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition IssuerAltName
char *decodeIssuerAltName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeGeneralNames(name,tag,tagOption,P,0,0,optional,Default,"2.5.29.18",xps_modtail, P_CRL );
	return S;
}
 // Definition _11
char *decode_11(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"",-1,0,0,(_CPFPLUS)decodeAttribute,0,0,""};
	TypeDescription _TD[] = {_TD0};

	S = decodeSEQUENCE_OF(name,tag,tagOption,P,1,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition SubjectDirectoryAttributes
char *decodeSubjectDirectoryAttributes(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decode_11(name,tag,tagOption,P,0,0,optional,Default,"2.5.29.9",xps_modtail, P_CRL );
	return S;
}
 // Definition BasicConstraints
char *decodeBasicConstraints(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"cA",-1,1,0,(_CPFPLUS)decodeBOOLEAN,0,1,""};
	TypeDescription _TD1 = {"pathLenConstraint",-1,2,0,(_CPFPLUS)decodeINTEGER,1,0,""};
	TypeDescription _TD[] = {_TD0,_TD1};

	S = decodeSEQUENCE(name,tag,tagOption,P,2,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition NameConstraints
char *decodeNameConstraints(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"permittedSubtrees",160,0,1,(_CPFPLUS)decodeGeneralSubtrees,1,0,""};
	TypeDescription _TD1 = {"excludedSubtrees",161,0,1,(_CPFPLUS)decodeGeneralSubtrees,1,0,""};
	TypeDescription _TD[] = {_TD0,_TD1};

	S = decodeSEQUENCE(name,tag,tagOption,P,2,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition _12
char *decode_12(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"",-1,0,0,(_CPFPLUS)decodeGeneralSubtree,0,0,""};
	TypeDescription _TD[] = {_TD0};

	S = decodeSEQUENCE_OF(name,tag,tagOption,P,1,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition GeneralSubtrees
char *decodeGeneralSubtrees(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decode_12(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition GeneralSubtree
char *decodeGeneralSubtree(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"base",-1,0,0,(_CPFPLUS)decodeGeneralName,0,0,""};
	TypeDescription _TD1 = {"minimum",160,0,1,(_CPFPLUS)decodeBaseDistance,0,1,""};
	TypeDescription _TD2 = {"maximum",161,0,1,(_CPFPLUS)decodeBaseDistance,1,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2};

	S = decodeSEQUENCE(name,tag,tagOption,P,3,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition BaseDistance
char *decodeBaseDistance(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeINTEGER(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition PolicyConstraints
char *decodePolicyConstraints(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"requireExplicitPolicy",160,0,1,(_CPFPLUS)decodeSkipCerts,1,0,""};
	TypeDescription _TD1 = {"inhibitPolicyMapping",161,0,1,(_CPFPLUS)decodeSkipCerts,1,0,""};
	TypeDescription _TD[] = {_TD0,_TD1};

	S = decodeSEQUENCE(name,tag,tagOption,P,2,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition SkipCerts
char *decodeSkipCerts(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeINTEGER(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition _13
char *decode_13(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"",-1,0,0,(_CPFPLUS)decodeDistributionPoint,0,0,""};
	TypeDescription _TD[] = {_TD0};

	S = decodeSEQUENCE_OF(name,tag,tagOption,P,1,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition CRLDistributionPoints
char *decodeCRLDistributionPoints(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decode_13(name,tag,tagOption,P,0,0,optional,Default,"2.5.29.31",xps_modtail, P_CRL );
	return S;
}
 // Definition DistributionPoint
char *decodeDistributionPoint(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"distributionPoint",160,0,1,(_CPFPLUS)decodeDistributionPointName,1,0,""};
	TypeDescription _TD1 = {"reasons",161,0,1,(_CPFPLUS)decodeReasonFlags,1,0,""};
	TypeDescription _TD2 = {"cRLIssuer",162,0,1,(_CPFPLUS)decodeGeneralNames,1,0,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2};

	S = decodeSEQUENCE(name,tag,tagOption,P,3,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition DistributionPointName
char *decodeDistributionPointName(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"fullName",160,0,1,(_CPFPLUS)decodeGeneralNames,0,0,""};
	TypeDescription _TD1 = {"nameRelativeToCRLIssuer",161,0,1,(_CPFPLUS)decodeRelativeDistinguishedName,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1};

	S = decodeCHOICE(name,tag,tagOption,P,2,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition ReasonFlags
char *decodeReasonFlags(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeBIT_STRING(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition _14
char *decode_14(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"",-1,0,0,(_CPFPLUS)decodeKeyPurposeId,0,0,""};
	TypeDescription _TD[] = {_TD0};

	S = decodeSEQUENCE_OF(name,tag,tagOption,P,1,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition ExtKeyUsageSyntax
char *decodeExtKeyUsageSyntax(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decode_14(name,tag,tagOption,P,0,0,optional,Default,"2.5.29.37",xps_modtail, P_CRL );
	return S;
}
 // Definition KeyPurposeId
char *decodeKeyPurposeId(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeOBJECT_IDENTIFIER(name,tag,tagOption,P,0,0,optional,Default,"",xps_modtail, P_CRL );
	return S;
}
 // Definition InhibitAnyPolicy
char *decodeInhibitAnyPolicy(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeSkipCerts(name,tag,tagOption,P,0,0,optional,Default,"2.5.29.54",xps_modtail, P_CRL );
	return S;
}
 // Definition FreshestCRL
char *decodeFreshestCRL(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeCRLDistributionPoints(name,tag,tagOption,P,0,0,optional,Default,"2.5.29.46",xps_modtail, P_CRL );
	return S;
}
 // Definition _15
char *decode_15(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"",-1,0,0,(_CPFPLUS)decodeAccessDescription,0,0,""};
	TypeDescription _TD[] = {_TD0};

	S = decodeSEQUENCE_OF(name,tag,tagOption,P,1,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition AuthorityInfoAccessSyntax
char *decodeAuthorityInfoAccessSyntax(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decode_15(name,tag,tagOption,P,0,0,optional,Default,"1.3.6.1.5.5.7.1.1",xps_modtail, P_CRL );
	return S;
}
 // Definition AccessDescription
char *decodeAccessDescription(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"accessMethod",-1,0,0,(_CPFPLUS)decodeOBJECT_IDENTIFIER,0,0,""};
	TypeDescription _TD1 = {"accessLocation",-1,0,0,(_CPFPLUS)decodeGeneralName,0,0,""};
	TypeDescription _TD[] = {_TD0,_TD1};

	S = decodeSEQUENCE(name,tag,tagOption,P,2,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition _16
char *decode_16(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"",-1,0,0,(_CPFPLUS)decodeAccessDescription,0,0,""};
	TypeDescription _TD[] = {_TD0};

	S = decodeSEQUENCE_OF(name,tag,tagOption,P,1,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition SubjectInfoAccessSyntax
char *decodeSubjectInfoAccessSyntax(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decode_16(name,tag,tagOption,P,0,0,optional,Default,"1.3.6.1.5.5.7.1.11",xps_modtail, P_CRL );
	return S;
}
 // Definition CRLNumber
char *decodeCRLNumber(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeINTEGER(name,tag,tagOption,P,0,0,optional,Default,"2.5.29.20",xps_modtail, P_CRL );
	return S;
}
 // Definition IssuingDistributionPoint
char *decodeIssuingDistributionPoint(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	TypeDescription _TD0 = {"distributionPoint",160,0,1,(_CPFPLUS)decodeDistributionPointName,1,0,""};
	TypeDescription _TD1 = {"onlyContainsUserCerts",129,1,1,(_CPFPLUS)decodeBOOLEAN,0,1,""};
	TypeDescription _TD2 = {"onlyContainsCACerts",130,1,1,(_CPFPLUS)decodeBOOLEAN,0,1,""};
	TypeDescription _TD3 = {"onlySomeReasons",163,0,1,(_CPFPLUS)decodeReasonFlags,1,0,""};
	TypeDescription _TD4 = {"indirectCRL",132,1,1,(_CPFPLUS)decodeBOOLEAN,0,1,""};
	TypeDescription _TD5 = {"onlyContainsAttributeCerts",133,1,1,(_CPFPLUS)decodeBOOLEAN,0,1,""};
	TypeDescription _TD[] = {_TD0,_TD1,_TD2,_TD3,_TD4,_TD5};

	S = decodeSEQUENCE(name,tag,tagOption,P,6,_TD,optional,Default,"", xps_modtail, P_CRL );
	return S;
}
 // Definition BaseCRLNumber
char *decodeBaseCRLNumber(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeCRLNumber(name,tag,tagOption,P,0,0,optional,Default,"2.5.29.27",xps_modtail, P_CRL );
	return S;
}
 // Definition CRLReason
char *decodeCRLReason(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeENUMERATED(name,tag,tagOption,P,0,0,optional,Default,"2.5.29.21",xps_modtail, P_CRL );
	return S;
}
 // Definition CertificateIssuer
char *decodeCertificateIssuer(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeGeneralNames(name,tag,tagOption,P,0,0,optional,Default,"2.5.29.29",xps_modtail, P_CRL );
	return S;
}
 // Definition HoldInstructionCode
char *decodeHoldInstructionCode(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeOBJECT_IDENTIFIER(name,tag,tagOption,P,0,0,optional,Default,"2.5.29.23",xps_modtail, P_CRL );
	return S;
}
 // Definition InvalidityDate
char *decodeInvalidityDate(char * name,int tag,int tagOption,byte **P,int count,TypeDescription TD[],int optional,int Default,char *ObjectIdentifier, Modifications ***xps_modtail, byte **P_CRL )
{
	char *S;
	S = decodeGeneralizedTime(name,tag,tagOption,P,0,0,optional,Default,"2.5.29.24",xps_modtail, P_CRL );
	return S;
}


int xps_init( Operation *xps_op, Operation *op )
{
	Connection *conn = op->o_conn;
	
	xps_op->o_req_dn = op->o_req_dn;
	xps_op->o_req_ndn = op->o_req_ndn;
	xps_op->o_bd = select_backend( &( xps_op->o_req_ndn ), 0, 1 );

	if ( xps_op->o_bd && xps_op->o_bd->be_search ) {
		/* init xps_op */
		xps_op->o_protocol = LDAP_VERSION3;
		xps_op->o_ndn = conn->c_ndn;
		xps_op->o_time = slap_get_time();
		xps_op->o_do_not_cache = 1;
		xps_op->o_is_auth_check = 0;
		xps_op->o_threadctx = op->o_threadctx;// conn->c_sasl_bindop->o_threadctx;
		xps_op->o_tmpmemctx = op->o_tmpmemctx;//conn->c_sasl_bindop->o_tmpmemctx;
		xps_op->o_tmpmfuncs = op->o_tmpmfuncs;//conn->c_sasl_bindop->o_tmpmfuncs;
		xps_op->o_conn = conn;
		xps_op->o_connid = conn->c_connid;
		return LDAP_SUCCESS;
	} else
		return LDAP_UNWILLING_TO_PERFORM;
}

int xps_create_entries(
	Operation *op,
	SlapReply *rs,
	Modifications	**modtail,
	Entry	***xps_e_array,
	Entry	***xps_e_revoked,
	struct berval dn )
{
	int x509_found = 0; /* flag, indicating that X.509 atrribute was found in the list of mods*/
	int found_type; 

	Debug(LDAP_DEBUG_TRACE, "===> xps_creat_enries: \n",0 , 0, 0);
	
	while ( *modtail != NULL )
	{
		struct xps_x509_type *xps_tail;
		int i;
		found_type = 0;
		Modifications *tmp = ( *modtail );
		
		xps_tail = pkcTypes;
		while ( xps_tail != NULL && !found_type )
		{
			for ( i = 0; xps_tail->attributeTypes[i] != NULL; i++ )
			{
				if ( strcasecmp( tmp->sml_type.bv_val, xps_tail->attributeTypes[i] ) == 0 ) {
					rs->sr_err = xps_x509attrs2entries( op, rs, tmp,
						xps_tail->oc, "x509PKC", DN_TYPE_CERT, dn,
						decodeCertificate, "Certificate" , xps_e_array, NULL );
					found_type = 1;
					x509_found = 1;
					break;
				}
			}
			xps_tail = xps_tail->next;
		}

		xps_tail = crlTypes;
		while ( xps_tail != NULL && !found_type )
		{
			for ( i = 0; xps_tail->attributeTypes[i] != NULL; i++ )
			{
				if ( strcasecmp( tmp->sml_type.bv_val, xps_tail->attributeTypes[i] ) == 0 ) {
					rs->sr_err = xps_x509attrs2entries( op, rs, tmp,
						xps_tail->oc, "x509CRL", DN_TYPE_CRL, dn,
						decodeCertificateList, "certificateList" , xps_e_array, xps_e_revoked);
					found_type = 1;
					x509_found = 1;
					break;
				}
			}
			xps_tail = xps_tail->next;
		}
		
		xps_tail = acTypes;
		while ( xps_tail != NULL && !found_type )
		{
			for ( i = 0; xps_tail->attributeTypes[i] != NULL; i++ )
			{
				if ( strcasecmp( tmp->sml_type.bv_val, xps_tail->attributeTypes[i] ) == 0 ) {
					rs->sr_err = xps_x509attrs2entries( op, rs, tmp,
						xps_tail->oc, "x509AC", DN_TYPE_CERT, dn,
						decodeAttributeCertificate, "AttributeCertificate" , xps_e_array, NULL);
					found_type = 1;
					x509_found = 1;
					break;
				}
			}
			xps_tail = xps_tail->next;
		}

		if ( !duplicateAttribute && found_type )
		{
			/* if the X.509 attribute is found and duplicateAttribute is false
			* - we don't keep this attr */
			(*modtail) = (*modtail)->sml_next;

			slap_mod_free( &tmp->sml_mod, 0 );
			free( tmp );
		} else {
			if ( (*modtail) != NULL )
				modtail = &(*modtail)->sml_next;
		}
	}
	Debug(LDAP_DEBUG_TRACE, "<=== xps_creat_enries: x509_found %d\n",x509_found , 0, 0);
	
	return x509_found;
}

/*Create an array of Entry structures from the x509 attribute*/
int xps_x509attrs2entries(
	Operation *op,
	SlapReply *rs,
	Modifications	*mod,
	char *oc_type,
	char *abs_oc_type,
	int dnflag,
	struct berval dn,
	_CPFPLUS decodeFUNCTION,
	char *decodeName,
	Entry ***e_pointer,
	Entry ***e_revoked )
{

	int rc, i, num = 0;
	Entry **e = *e_pointer;
	int repl_user = 0;
	int update = 0;
	char *res = NULL;
	char textbuf[SLAP_TEXT_BUFLEN];
	size_t textlen = sizeof( textbuf );
	const char *text;

	byte *P, *P_CRL = NULL;
	struct berval xps_dn = { 0, NULL };

	Modifications	*tmp = NULL;
	Modifications	*xps_modlist = NULL;
	Modifications	**xps_modtail = &xps_modlist;
	
	int tag;

Debug(LDAP_DEBUG_TRACE, "===> xps_creat_enries: \n",0 , 0, 0);

	// count the number of the already created entries
	if ( e != NULL ) {
		for ( num = 0; e[num] != NULL; num++ );
	}

	// count the number of the attr values
	for ( i = 0; mod->sml_values[i].bv_val != NULL; i++ );
	
	e = (Entry **)ch_realloc( e, sizeof( Entry *) * ( i + num + 1 ) );
	e[i + num ] = NULL;

	for ( i = 0; mod->sml_values[i].bv_val != NULL; i++ )
	{
		xps_add_oc( oc_type, abs_oc_type, &xps_modtail, &mod->sml_type, &mod->sml_values[i] );
		
		P = ( byte *)mod->sml_values[i].bv_val;
		
		res = decodeFUNCTION(decodeName,0,0,&P,0,0,0,0, NULL, &xps_modtail, &P_CRL);
		if( res == NULL ) {
			if ( xps_modlist != NULL ) {
				slap_mods_free( xps_modlist );
			}
			Debug( LDAP_DEBUG_ANY," decodeFUNCTION failed\n", 0, 0, 0 );
			return ( LDAP_UNWILLING_TO_PERFORM );
		}

		e[i + num] = (Entry *) ch_calloc( 1, sizeof(Entry) );

		rc = xps_create_dn( dnflag, &dn, &xps_dn, xps_modlist);
		if (rc != LDAP_SUCCESS ) {
			if ( xps_modlist != NULL ) {
				slap_mods_free( xps_modlist );
			}
			Debug( LDAP_DEBUG_ANY," xps_create_dn failed\n", 0, 0, 0 );
			return ( LDAP_UNWILLING_TO_PERFORM );
		}

		rc = dnPrettyNormal( NULL, &xps_dn, &e[i + num]->e_name,
			&e[i + num]->e_nname, op->o_tmpmemctx );
		if (rc != LDAP_SUCCESS ) {
			if ( xps_modlist != NULL ) {
				slap_mods_free( xps_modlist );
			}
			Debug( LDAP_DEBUG_ANY," dnPrettyNormal failed\n", 0, 0, 0 );
			return ( LDAP_UNWILLING_TO_PERFORM );
		}

		Debug( LDAP_DEBUG_ANY," start slap_mods_check\n", 0, 0, 0 );
		rs->sr_err = slap_mods_check( xps_modlist, 0, &rs->sr_text,
										  textbuf, textlen, NULL );
		if (rc != LDAP_SUCCESS ) {
			if ( xps_modlist != NULL ) {
				slap_mods_free( xps_modlist );
			}
			Debug( LDAP_DEBUG_ANY," slap_mods_check failed\n", 0, 0, 0 );
			return ( LDAP_UNWILLING_TO_PERFORM );
		}

		/* TODO: fixme */
		tag = op->o_tag;
		op->o_tag = LDAP_REQ_ADD;
		Debug( LDAP_DEBUG_ANY," start slap_mods_opattrs\n", 0, 0, 0 );
		rs->sr_err = slap_mods_opattrs( op, xps_modlist, xps_modtail,
				&rs->sr_text, textbuf, textlen );
		
		op->o_tag = tag;
		
		if (rc != LDAP_SUCCESS ) {
			if ( xps_modlist != NULL ) {
				slap_mods_free( xps_modlist );
			}
			Debug( LDAP_DEBUG_ANY," slap_mods_opattrs failed\n", 0, 0, 0 );
			return ( LDAP_UNWILLING_TO_PERFORM );
		}

		rs->sr_err = slap_mods2entry( xps_modlist, &( e[num + i] ), repl_user, 0,
							&rs->sr_text, textbuf, textlen );
		if (rc != LDAP_SUCCESS ) {
			if ( xps_modlist != NULL ) {
				slap_mods_free( xps_modlist );
			}
			Debug( LDAP_DEBUG_ANY," slap_mods2entry failed\n", 0, 0, 0 );
			return ( LDAP_UNWILLING_TO_PERFORM );
		}

		/* free modlist for future usage */
		if ( xps_modlist != NULL ) {
			slap_mods_free( xps_modlist );
		}
		xps_modlist = NULL;
		xps_modtail = &xps_modlist;

		if (P_CRL != NULL )
		{
			int i, length;
			byte *T;
			int tlvTag;
			int bytes_decoded = 0;
			byte *start;
			_CPFPLUS F;
			char *s;
			char *returnString, *tempString;
			int n = 0;
			struct berval dn_revoked = { 0, NULL };

			length = decodeTL (&P_CRL);
			if ( length == -1 )
				return LDAP_UNWILLING_TO_PERFORM;
			
			start = P_CRL;
			returnString = newString( "" );

			if ( e_revoked != NULL && *e_revoked != NULL ) {
				for ( n = 0; ( *e_revoked )[n] != NULL; n++ );
			}

			while ((i = ((int) P_CRL - (int) start)) < length)
			{
				*e_revoked = (Entry **)ch_realloc( *e_revoked, sizeof( Entry *) * ( n + 2 ) );
				( *e_revoked )[n + 1] = NULL;
				n++;

				xps_add_oc(  "x509CRLentry" , NULL, &xps_modtail, NULL, NULL );
				
				tempString =
					 decode_5_1("revokedCertificate" , -1, 0, &P_CRL,
					0, 0, 0, 0, "", &xps_modtail, NULL );
	
				if ( tempString == NULL )	// an error
					return LDAP_UNWILLING_TO_PERFORM;
				free (tempString);

				rc = xps_create_dn( DN_TYPE_REVOKED, &xps_dn, &dn_revoked, xps_modlist);

				 (*e_revoked )[n-1] = (Entry *) ch_calloc( 1, sizeof(Entry) );
				
				rc = dnPrettyNormal( NULL, &dn_revoked, &( (*e_revoked )[n-1]->e_name ),
					&( (*e_revoked )[n-1]->e_nname ), op->o_tmpmemctx );

				rs->sr_err = slap_mods_check( xps_modlist, 0, &rs->sr_text,
					textbuf, textlen, NULL );
				
				/* TODO: fixme */
				tag = op->o_tag;
				op->o_tag = LDAP_REQ_ADD;
				Debug( LDAP_DEBUG_ANY," start slap_mods_opattrs\n", 0, 0, 0 );
				rs->sr_err = slap_mods_opattrs( op, xps_modlist, xps_modtail,
						&rs->sr_text, textbuf, textlen );
				op->o_tag = tag;

				rs->sr_err = slap_mods2entry( xps_modlist, &( (*e_revoked )[n-1] ), repl_user, 0,
					&rs->sr_text, textbuf, textlen );

				/* free modlist for future usage */
				if ( xps_modlist != NULL ) {
					slap_mods_free( xps_modlist );
				}
				xps_modlist = NULL;
				xps_modtail = &xps_modlist;
			}
			P_CRL =NULL;
		}
	}
	
	*e_pointer = e;
	Debug(LDAP_DEBUG_TRACE, "<=== xps_creat_enries \n",0 , 0, 0);
	return LDAP_SUCCESS;
}

/* open Write Ahead Log. It stored the info for rolling back the*
 * XPS operation if it wasn't successfully completed */
int xps_open_wal( FILE **wal, char **walname_p, struct berval dn )
{
	MD5_CTX c;
	unsigned char checksum[MD5_DIGEST_LENGTH ];
	int i, pathlen, rc = 0;
	char *fname;
	
Debug(LDAP_DEBUG_TRACE, "===> xps_open_wal: \n",0 , 0, 0);
	ldap_pvt_thread_mutex_lock( &xps_mutex );

	/* create a hash  of the provided DN string - it will be used as
	 * a part of the filename */
	MD5_Init (&c);
	MD5_Update (&c, dn.bv_val, dn.bv_len);
	MD5_Final (checksum, &c);

	/*fname is "WAL path / constructed WAL file name" */
	pathlen = strlen( WALpath );
	fname = (char *) ch_calloc( pathlen + 1 + 40, sizeof(char) );
	fname = strcpy( fname, WALpath );
	fname = strcat( fname, "/wal" );
	
	for (i = 0; i < MD5_DIGEST_LENGTH; i++)
	{
		sprintf (&(fname[pathlen + 4 +2*i]),"%02x", (unsigned int) checksum[i]);
	}
	sprintf (&(fname[pathlen + 36]), ".log");
	*walname_p = fname;
	
Debug(LDAP_DEBUG_TRACE, "<=== xps_open_wal: walname=%s\n",*walname_p , 0, 0);
	*wal = fopen( fname, "w+" );
	if ( *wal != NULL ) {
		Debug(LDAP_DEBUG_TRACE, "<=== xps_open_wal: LDAP_SUCCESS \n",0 , 0, 0);
		return LDAP_SUCCESS;
	}	else {
		Debug(LDAP_DEBUG_TRACE, "<=== xps_open_wal: LDAP_UNWILLING_TO_PERFORM \n",0 , 0, 0);
		ldap_pvt_thread_mutex_unlock( &xps_mutex );
		return LDAP_UNWILLING_TO_PERFORM;
	}
	flockfile ( *wal );
}

int xps_close_wal( FILE *wal, char *walname )
{
	Debug(LDAP_DEBUG_TRACE, "===> xps_close_wal: %s\n",walname, 0, 0);

	if ( wal ) {
		/* make this file accessible for other theads again */
		funlockfile( wal );
		/* close WAL file */
		fclose( wal );
		/* and delete it */
		unlink ( walname);

		ldap_pvt_thread_mutex_unlock( &xps_mutex );
	}
	
	if ( walname != NULL )
		free( walname );
	
Debug(LDAP_DEBUG_TRACE, "<=== xps_close_wal: \n",0 , 0, 0);
	return LDAP_SUCCESS;
}
	
int xps_dn2wal(	FILE *wal, struct berval *dn )
{
Debug(LDAP_DEBUG_TRACE, "===> xps_dn2wal: \n",0 , 0, 0);
	if ( dn->bv_val == NULL ) {
		Debug(LDAP_DEBUG_TRACE, "<=== xps_dn2wal: \n",0 , 0, 0);
		return LDAP_UNWILLING_TO_PERFORM;
	} else {
		Debug( LDAP_DEBUG_ANY,"xps_dn2wal: >>>>>>>>dn->bv_val = %s \n", dn->bv_val, 0, 0 );
		fprintf( wal, "\ndn: %s\n", dn->bv_val );
		fprintf( wal, "changetype: delete\n" );
		fflush( wal );
Debug(LDAP_DEBUG_TRACE, "<=== xps_dn2wal: \n",0 , 0, 0);
		return LDAP_SUCCESS;
	}
}

int
xps_entry2wal(	FILE *wal, Entry *e )
{
	char *data;
	int len;
	
Debug(LDAP_DEBUG_TRACE, "===> xps_entry2wal: \n",0 , 0, 0);
	data = entry2str( e, &len );
	fputs( "\n", wal );
	fputs( data, wal );
	fflush( wal );
Debug(LDAP_DEBUG_TRACE, "<=== xps_entry2wal: \n",0 , 0, 0);
	return LDAP_SUCCESS;
}

int xps_process_wal( Operation *op, SlapReply *rs, FILE *wal, char *walname )
{
	FILE *xps_log;
	char *data, *buf, *line, *type, *dnline;
	int len, rc = LDAP_SUCCESS;
	struct berval val;
	Entry *e;
	
	xps_log = fopen( XPSerrorlog, "a" );
	Debug( LDAP_DEBUG_ANY, "===> xps_process_wal\n", 0, 0, 0 );	
	data = xps_read_last_record( wal );
	Debug( LDAP_DEBUG_ANY, "xps_process_wal: data \n%s\n", data, 0, 0 );	

	while ( data != NULL )
	{		
		
		buf = ch_strdup( data );

		dnline = ldif_getline( &data );
		line = ldif_getline( &data );
	
		ldif_parse_line( line, &type, &val.bv_val, &val.bv_len );
	
		if ( strcasecmp( val.bv_val, "delete" )==0 &&  strcasecmp( type, "changetype" )==0  )
		{
			slap_callback cb = { NULL, slap_null_cb, NULL, NULL };
			ber_memfree( type );
			ber_memfree( val.bv_val );
			ldif_parse_line( dnline, &type, &val.bv_val, &val.bv_len );
			Debug( LDAP_DEBUG_ANY, "xps_process_wal: DELETE type:%s value:%s\n", type, val.bv_val, 0 );	

			op->o_tag = LDAP_REQ_ADD;
			op->ors_scope = 0;
			op->ors_deref = 0;
			op->ors_filter = NULL;
			op->o_callback = &cb;
			
			op->o_req_dn = val;
			op->o_req_ndn = val;

			op->o_bd->be_delete( op, rs );
			if ( rs->sr_err != LDAP_SUCCESS ) {
				rc = LDAP_UNWILLING_TO_PERFORM;
				fprintf( xps_log, "#\tXPS was unable to delete entry:\ndn: %s\n\t#reason: %s\n" ,  val.bv_val, rs->sr_text );
				fflush( xps_log );
			}
		} else {
			e = str2entry( buf );
			op->o_tag = LDAP_REQ_ADD;
			op->ors_scope = 0;
			op->ors_deref = 0;
			op->ors_filter = NULL;
			op->ora_e = e;

			if ( (op->o_bd->be_add)( op, rs ) == 0 ) {
				be_entry_release_w( op, e );
				e = NULL;
			} else {
				rc = LDAP_UNWILLING_TO_PERFORM;
				fprintf( xps_log, "#XPS was unable to add entry:\n%s\n#reason: %s\n\n" ,  entry2str( e, &len), rs->sr_text );
				fflush( xps_log );
				entry_free( e );
			}
		}
		rs->sr_err = xps_delete_last_record( wal, walname );

		if ( buf ){
			free( buf );
		}
		if ( type )
			ber_memfree( type );
		if ( val.bv_val )
			ber_memfree( val.bv_val );
		Debug( LDAP_DEBUG_ANY, "xps_process_wal: Trying to read last record...\n", 0, 0, 0 );	
		data = xps_read_last_record( wal );
		Debug( LDAP_DEBUG_ANY, "xps_process_wal: Last record in WAL:\n%s \n", data, 0, 0 );	
	}
	
	if ( xps_log )
		fclose( xps_log );
	
	Debug( LDAP_DEBUG_ANY, "<=== xps_process_wal\n", 0, 0, 0 );	
	return rc;
}

/*Add entries listed in "Entry **e" to the Directory*/
int xps_add_entries( Operation *op, SlapReply *rs, FILE *wal, Entry **e, char *walname )
{
	int i, failed = 0;
	int j=0;

Debug(LDAP_DEBUG_TRACE, "===> xps_add_entries: \n",0 , 0, 0);
	if (e != NULL )
	{
		for (i = 0; e[i] != NULL; i++ )
		{
Debug(LDAP_DEBUG_TRACE, "===> xps_add_entries:	op->o_req_dn = e[i]->e_nname %s\n",	e[i]->e_nname.bv_val , 0, 0);
			/* if adding is failed at some point - continue the 
				loop just to free the rest of the entries */
			if ( !failed )
			{
				xps_dn2wal( wal, &e[i]->e_nname );
				op->ora_e = e[i];
				op->o_req_dn = e[i]->e_nname;
				op->o_req_ndn = e[i]->e_nname;

				if ( (op->o_bd->be_add)( op, rs ) == 0 ) {
					be_entry_release_w( op, e[i] );
					e[i] = NULL;
				} else {
					xps_read_last_record( wal );
					xps_delete_last_record( wal, walname );
					Debug(LDAP_DEBUG_TRACE, "===> xps_add_entries: failed to add: %s\n",e[i]->e_nname.bv_val  , 0, 0);
					entry_free( e[i] );
					failed = 1;
				}
				
			} else {
				entry_free( e[i] );
			}
		}
		free( e );
	}
Debug(LDAP_DEBUG_TRACE, "<=== xps_add_entries: \n",0 , 0, 0);
		
	if ( failed ) {
		return LDAP_UNWILLING_TO_PERFORM;
	} else {
		return LDAP_SUCCESS;
	}
}

/*Add object class and x509 attributes to the modifications list*/
void xps_add_oc (
	char	*oc,
	char *abs_oc,
	Modifications	***modtail_p,
	struct berval	*x509_type,
	struct berval	*x509_value )
{
	Modifications *xps_mod;
	struct berval type;
	struct berval type2;
	struct berval tmp_base;
	struct berval tmp_oc;
	struct berval tmp_abs_oc = { 0, NULL };
	struct berval tmp_null = { 0, NULL };
	BerVarray values;
	BerVarray values2;

Debug(LDAP_DEBUG_TRACE, "===> xps_add_oc: \n",0 , 0, 0);
	/* add objectClass attribute */
	tmp_base.bv_len = sizeof ("x509base")-1;
	tmp_base.bv_val = ch_strdup ("x509base");
	if (abs_oc != NULL ) {
		tmp_abs_oc.bv_len = strlen( abs_oc );
		tmp_abs_oc.bv_val = ch_strdup ( abs_oc );
	}
	tmp_oc.bv_len = strlen( oc );
	tmp_oc.bv_val = ch_strdup ( oc );

	type.bv_len = sizeof ("objectClass")-1;
	type.bv_val = ch_strdup ("objectClass");
 
	values = (BerVarray) ch_malloc( sizeof( struct berval ) * 4 );
	values[0] = tmp_base;
	values[1] = tmp_oc; 
	if (tmp_abs_oc.bv_val != NULL ) {
		values[2] = tmp_abs_oc;
		values[3] = tmp_null;
	} else 
		values[2] = tmp_null;
	
	xps_mod  = (Modifications *) ch_malloc( sizeof(Modifications) );
	
 	xps_mod->sml_op = LDAP_MOD_ADD;
	xps_mod->sml_next = NULL;
	xps_mod->sml_desc = NULL;
	xps_mod->sml_type = type;
	xps_mod->sml_values = values;
	xps_mod->sml_nvalues = NULL;
	

	*( *modtail_p ) = xps_mod;
	*modtail_p = &xps_mod->sml_next;
	
	/* add X.509 attribute */
	if ( x509_type != NULL ) {
		
		ber_dupbv( &type2, x509_type );
		
		values2 = (BerVarray) ch_malloc( sizeof( struct berval ) * 2 );
		ber_dupbv( &values2[0], x509_value );
		ber_dupbv( &values2[1], &tmp_null );
		
		xps_mod  = (Modifications *) ch_malloc( sizeof(Modifications) );
		
		xps_mod->sml_op = LDAP_MOD_ADD;
		xps_mod->sml_next = NULL;
		xps_mod->sml_desc = NULL;
		xps_mod->sml_type = type2;
		xps_mod->sml_values = values2;
		xps_mod->sml_nvalues = NULL;
	
		*( *modtail_p ) = xps_mod;
		*modtail_p = &xps_mod->sml_next;
	}

	Debug(LDAP_DEBUG_TRACE, "<=== xps_add_oc: \n",0 , 0, 0);
}

/* 
 * create DN for the X.509 entry according to the conf option:
 * "x509crlThisUpdate,baseDN" or
 * "x509serialNumber,baseDN" or
 * "x509serialNumber$x509issuer,baseDN" or
 * "x509issuerSerial,baseDN"
 */
int xps_create_dn(
	int				dnflag,
	struct berval	*dn,
	struct berval	*xps_dn,
	Modifications	*xps_modlist )
{
	char * value, *value2;
	int format;
	
	switch ( dnflag ) {
		case DN_TYPE_CRL: // CRL: x509crlThisUpdate=<value>
			value = xps_find_attr( "x509crlThisUpdate", xps_modlist);
			if ( value == NULL ) 
				return LDAP_UNWILLING_TO_PERFORM;
			xps_dn->bv_len = sizeof( "x509crlThisUpdate=" ) + strlen( value ) + dn->bv_len;
			xps_dn->bv_val = (char *)ch_malloc( sizeof(char) * ( xps_dn->bv_len  + 1 ) );
			xps_dn->bv_val = strcpy( xps_dn->bv_val, "x509crlThisUpdate=" );
			xps_dn->bv_val = strcat (xps_dn->bv_val, value );
			xps_dn->bv_val = strcat (xps_dn->bv_val, ",");
			xps_dn->bv_val = strcat (xps_dn->bv_val, dn->bv_val);
			return LDAP_SUCCESS;
		case DN_TYPE_CERT: // AC or PKC
			format = certRDNformat;
			break;
		case DN_TYPE_REVOKED:// Revoked Certificate
			format = revokedRDNformat;
			break;
	}

	value = xps_find_attr( "x509serialNumber", xps_modlist);
	if ( value == NULL ) 
		return LDAP_UNWILLING_TO_PERFORM;
	
	if ( format == XPS_RDNFORMAT_X509SERIAL )
	{ /* x509serialNumber=<value>,<base DN> */
		xps_dn->bv_len = sizeof( "x509serialNumber=" ) + strlen( value ) + dn->bv_len;
		xps_dn->bv_val = (char *)ch_malloc( sizeof(char) * ( xps_dn->bv_len  + 1 ) );
		xps_dn->bv_val = strcpy( xps_dn->bv_val, "x509serialNumber=" );
		xps_dn->bv_val = strcat (xps_dn->bv_val, value );
	} else if ( format == XPS_RDNFORMAT_X509SERIAL_x509ISSUER ) {
		/* x509serialNumber=<value>+x509issuer=<value2>,<base DN> */
		value2 = xps_find_attr( "x509issuer", xps_modlist);
		if ( value == NULL ) 
			return LDAP_UNWILLING_TO_PERFORM;
		xps_dn->bv_len = strlen( value ) + strlen( value2 ) + dn->bv_len + 
			sizeof( "x509serialNumber=") + sizeof( "x509issuer=");
		xps_dn->bv_val = (char *)ch_malloc( sizeof(char) * ( xps_dn->bv_len  + 1 ) );
		xps_dn->bv_val = strcpy( xps_dn->bv_val, "x509serialNumber=" );
		xps_dn->bv_val = strcat (xps_dn->bv_val, value );
		xps_dn->bv_val = strcat (xps_dn->bv_val, "+");
		xps_dn->bv_val = strcat( xps_dn->bv_val, value2 );
	} else if ( format == XPS_RDNFORMAT_X509ISSUERSERIAL ) {
		/* x509issuerSerial=<value>$<value2>,<base DN> */
		value2 = xps_find_attr( "x509issuer", xps_modlist);
		if ( value2 == NULL ) 
			return LDAP_UNWILLING_TO_PERFORM;
		xps_dn->bv_len = sizeof( "x509issuerSerial=" ) + strlen( value ) + strlen( value2 ) + dn->bv_len + 2;
		xps_dn->bv_val = (char *)ch_malloc( sizeof(char) * ( xps_dn->bv_len  + 1 ) );
		xps_dn->bv_val = strcpy( xps_dn->bv_val, "x509issuerSerial=" );
		xps_dn->bv_val = strcat (xps_dn->bv_val, value );
		xps_dn->bv_val = strcat (xps_dn->bv_val, "$");
		xps_dn->bv_val = strcat( xps_dn->bv_val, value2 );
	}
	xps_dn->bv_val = strcat (xps_dn->bv_val, ",");
	xps_dn->bv_val = strcat (xps_dn->bv_val, dn->bv_val);

	return LDAP_SUCCESS;
}

char *xps_find_attr(
	char *type, 
	Modifications	*modlist )
{
	char *tmp = NULL;
	Modifications **xps_modtail = &modlist;
	
	for( ;*xps_modtail != NULL; xps_modtail = &(*xps_modtail)->sml_next )
	{
		Debug(LDAP_DEBUG_TRACE, "= xps_find_attr: %s %s\n", (*xps_modtail)->sml_type.bv_val, type, 0);
		if ( strcasecmp( (*xps_modtail)->sml_type.bv_val, type ) == 0 )
		{
			return strdup( (*xps_modtail)->sml_values[0].bv_val );
		}
	}
	return NULL;
}


int xps_add_mod( char *name, char *value, int len, Modifications ***xps_modtail_p, int flag )
{
	Modifications *xps_mod;
	struct berval type = { 0, NULL };
	BerVarray tmp_values;
	struct berval val = { 0, NULL };
	struct berval nul_val = { 0, NULL };
	struct xps_attr_type *attr = xpsattr;

	if ( flag )
		Debug(LDAP_DEBUG_TRACE, "= xps_add_mod: %s %s\n", name, value, 0);
	else
		Debug(LDAP_DEBUG_TRACE, "= xps_add_mod: %s %d bytes\n", name, len, 0);
	/* find matching X.509 attribute in xpsattr list */
	while ( attr != NULL )
	{
		if ( strcasecmp( name, attr->asn1str ) == 0 ) {
			if ( flag == 1 ) {
				type.bv_len = strlen( attr->xpsstr );
				type.bv_val = ch_strdup(attr->xpsstr);
				
				val.bv_len = len;
				val.bv_val = ch_strdup( value );

			} else {
				type.bv_len = strlen( attr->xpsstr ) + sizeof( ";binary" );
				type.bv_val = ch_strdup( attr->xpsstr );
				type.bv_val = ( char * )realloc( type.bv_val, type.bv_len + 1 );
				strcat( type.bv_val, ";binary" );

				val.bv_len = len;
				val.bv_val = ( char * )malloc( sizeof( char ) * len );
				val.bv_val = memcpy( val.bv_val, value, len );
			}

			tmp_values = ( BerVarray )ch_malloc( sizeof( struct berval ) * 2 );
			tmp_values[0] = val;
			tmp_values[1] = nul_val;

			xps_mod  = (Modifications *) ch_malloc( sizeof(Modifications) );
			
			xps_mod->sml_op = LDAP_MOD_ADD;
			xps_mod->sml_next = NULL;
			xps_mod->sml_desc = NULL;
			xps_mod->sml_type = type;
			xps_mod->sml_values = tmp_values;
			xps_mod->sml_nvalues = NULL;
		
			**xps_modtail_p = xps_mod;
			*xps_modtail_p = &xps_mod->sml_next;
			break;
			
		} else {
			attr = attr->next;
		}
	}
	return 0;
}

int xps_free_entries( Operation *o, Entry **e )
{
	int i, num;
Debug(LDAP_DEBUG_TRACE, "===> xps_free_entries:  %d\n",e , 0, 0);
	
	if ( e != NULL )
	{
		for ( i = 0; e[i] != NULL; i++ )
		{
			entry_free( e[i] );
		}
		free( e );
	}
	return LDAP_SUCCESS;
}

/*Callback function: instead of sending result to the client put it*/
/*into the appropriate array xps_e_level2 or xps_e_level3 */
int
xps_send_search_entry( Operation *op, SlapReply *rs )
{
	int rc, i;

	AttributeDescription *ad;
	const char *text;
	Attribute *a;
	BerVarray bv;
	char *buf;
	int num, len, oc_found = 0;
	Entry *e = rs->sr_entry;
	struct xps_x509_type *tail;
	
	if (rs->sr_type != REP_SEARCH) return 0;

	if ( rs->sr_entry != NULL )
		e = rs->sr_entry;
	else return 0;
	
Debug(LDAP_DEBUG_TRACE, "===> xps_send_search_entry:  dn = %s\n",rs->sr_entry->e_name.bv_val, 0, 0);
	ad = NULL;
	slap_str2ad( "objectClass", &ad, &text );
	
	a = attr_find( e->e_attrs, ad );
	
	if ( a->a_vals != NULL )
	{
		for ( bv = a->a_vals; bv->bv_val; bv++ ) {
			/*TODO: fixme */
			if ( strcasecmp( bv->bv_val, "x509PKC" ) == 0 ||
				strcasecmp( bv->bv_val, "x509AC" )  == 0||
				strcasecmp( bv->bv_val, "x509CRL" )  == 0 )
			{
				/* add entry to level 2 */
				if ( xps_e_level2 != NULL )
				{
					for ( num = 0; xps_e_level2[num] != NULL; num++ );
				} else
					num = 0;
				
				xps_e_level2 = (Entry **)ch_realloc( xps_e_level2, 
					sizeof( Entry *) * ( num + 2 ) );
				xps_e_level2[num] = entry_dup( e );
				xps_e_level2[num + 1] = NULL;
				break;
			}

			if ( strcasecmp( bv->bv_val, "x509CRLentry" )  == 0 ){
				/* add entry to level 3 */
				if ( xps_e_level3 != NULL )	
				{
					for ( num = 0; xps_e_level3[num] != NULL; num++ );
				} else
					num = 0;
				
				xps_e_level3 = (Entry **)ch_realloc( xps_e_level3, 
					sizeof( Entry *) * ( num + 2 ) );
				xps_e_level3[num] = entry_dup( e );
				xps_e_level3[num + 1] = NULL;
				break;
			}
			
		}
	}
	Debug(LDAP_DEBUG_TRACE, "<=== xps_send_search_entry dn = %s\n",0, 0, 0);
	return LDAP_SUCCESS;
}

/* creat a PRESENT filter if a_value is empty or an */
/* EQUALITY filter if opposite */
int xps_create_filter(
Operation *op, 
Filter **filt,
struct berval *a_type,
struct berval *a_value)
{
	struct berval type, value;
	AttributeAssertion *aa;
	const char **text;
	Filter f;
	int rc = LDAP_SUCCESS;
	
Debug(LDAP_DEBUG_TRACE, "===> xps_create_filter: \n",0 , 0, 0);
	
	f.f_next = NULL;
	type.bv_val = strdup( a_type->bv_val );
	type.bv_len = a_type->bv_len;

	if ( a_value != NULL )
	{	/* equality filter */
		ber_dupbv( &value, a_value );
	
		f.f_choice = LDAP_FILTER_EQUALITY;

		aa = op->o_tmpalloc( sizeof( AttributeAssertion ), op->o_tmpmemctx );
		aa->aa_desc = NULL;
		aa->aa_value.bv_val = NULL;

		rc = slap_bv2ad( &type, &aa->aa_desc, text );
	
		if( rc != LDAP_SUCCESS ) {
		Debug( LDAP_DEBUG_FILTER,
			"----- xps_create_filter:  unknown attributeType %s\n", type.bv_val, 0, 0 );
			op->o_tmpfree( aa, op->o_tmpmemctx );
			return rc;
		}
		rc = asserted_value_validate_normalize(
			aa->aa_desc, ad_mr(aa->aa_desc, SLAP_MR_EQUALITY),
			SLAP_MR_EQUALITY, &value, &aa->aa_value, text, op->o_tmpmemctx );

		if( rc != LDAP_SUCCESS ) {
			Debug( LDAP_DEBUG_FILTER,
			"----- xps_create_filter: illegal value for attributeType %s\n", type.bv_val, 0, 0 );
			op->o_tmpfree( aa, op->o_tmpmemctx );
			return rc;
		}
		
		f.f_ava = aa;
		assert( f.f_ava != NULL );
		
	} else { /* present filter */
		f.f_choice = LDAP_FILTER_PRESENT;
		f.f_desc = NULL;
		rc = slap_bv2ad( &type, &f.f_desc, text );
	}
	if ( rc == LDAP_SUCCESS ) {
		*filt = op->o_tmpalloc( sizeof(f), op->o_tmpmemctx );
		**filt = f;
		
	}
Debug(LDAP_DEBUG_TRACE, "<=== xps_create_filter: \n",0 , 0, 0);
	return rc;
}

/*Delete entries listed in xps_e_level2 and xps_e_level3 */
/* from the Directory*/
int xps_delete_entries ( Operation *op, SlapReply *rs, FILE *wal, char *walname )
{
	int rc = LDAP_SUCCESS;
	int i;
Debug(LDAP_DEBUG_TRACE, "===> xps_delete_entries: \n",0 , 0, 0);

	if ( xps_e_level3 != NULL )
	{
		for ( i = 0; xps_e_level3[i]	!= NULL; i++ )
		{
			xps_entry2wal( wal, xps_e_level3[i] );
			op->o_req_dn = xps_e_level3[i]->e_name;
			op->o_req_ndn = xps_e_level3[i]->e_nname;
	
			op->o_bd->be_delete( op, rs );
			
			if ( rs->sr_err != LDAP_SUCCESS ) {
Debug(LDAP_DEBUG_TRACE, "----------------- xps_delete_entries: deleting xps_e_level3 failed\n",0 , 0, 0);
				xps_read_last_record( wal );
				xps_delete_last_record( wal, walname );
				xps_process_wal( op, rs, wal, walname );
				return LDAP_UNWILLING_TO_PERFORM;
			}
		}
	}

	if ( xps_e_level2 != NULL )
	{
		for ( i = 0; xps_e_level2[i]	!= NULL; i++ )
		{
			xps_entry2wal( wal, xps_e_level2[i] );
			op->o_req_dn = xps_e_level2[i]->e_name;
			op->o_req_ndn = xps_e_level2[i]->e_nname;
	
			op->o_bd->be_delete( op, rs );
			
			if ( rs->sr_err != LDAP_SUCCESS ) {
Debug(LDAP_DEBUG_TRACE, "----------------- xps_delete_entries: deleting xps_e_level2 failed\n",0 , 0, 0);
				xps_read_last_record( wal );
				xps_delete_last_record( wal, walname );
				xps_process_wal( op, rs, wal, walname );
				return LDAP_UNWILLING_TO_PERFORM;
			}
		}
	}
	/*testing the WAL after successfully deleting all entries */
//	xps_process_wal( op, rs, wal, walname );
//	return LDAP_UNWILLING_TO_PERFORM;
	
Debug(LDAP_DEBUG_TRACE, "<=== xps_delete_entries: \n",0 , 0, 0);
	return LDAP_SUCCESS;
}


char *
xps_read_last_record( FILE *wal )
{
	char		*buf, line[ LDAPMOD_MAXLINE ];
	int		lcur, lmax;
	fpos_t	lastrec;

Debug(LDAP_DEBUG_TRACE, "===> xps_read_last_record: \n",0 , 0, 0);
    lcur = lmax = 0;
    buf = NULL;

	/* rewind */
	fseek(wal, 0L, SEEK_SET);

	fgetpos( wal, &lastrec );
	
    while ( fgets( line, sizeof(line), wal ) != NULL ) {
    	int len = strlen( line );

		if( len < 2 || ( len == 2 && *line == '\r' )) {
			fgetpos( wal, &lastrec );
		}
	}

	fsetpos( wal, &lastrec );

    while ( fgets( line, sizeof(line), wal ) != NULL ) {
    	int len = strlen( line );
		if( len < 2 || ( len == 2 && *line == '\r' )) {
			if( buf == NULL ) {
				continue;
			} else {
				break;
			}
		}
		if ( lcur + len + 1 > lmax ) {
			lmax = LDAPMOD_MAXLINE
				* (( lcur + len + 1 ) / LDAPMOD_MAXLINE + 1 );

			if (( buf = (char *)ber_memrealloc( buf, lmax )) == NULL ) {
				perror( "realloc" );
				exit( EXIT_FAILURE );
			}
		}

		strcpy( buf + lcur, line );
		lcur += len;
    }
	fsetpos( wal, &lastrec );

Debug(LDAP_DEBUG_TRACE, "<=== xps_read_last_record: \n",0 , 0, 0);
    return( buf );
}

int
xps_delete_last_record ( FILE *wal, char *walname )
{
	fpos_t ptr;
	int c;

	/*move one character back */
	ungetc( c, wal );
	/*get an offset */
	fgetpos( wal, &ptr );
	/* truncate file up to current offset */
	truncate( walname, ptr );
	fflush( wal );
	return LDAP_SUCCESS;
}

int xps_create_entries_for_modify(
	Operation *op,
	SlapReply *rs,
	Modifications	**modtail,
	Entry	***xps_e_array,
	Entry	***xps_e_revoked,
	struct berval dn )
{
	int x509_found = 0; /* flag, indicating that X.509 atrribute was found in the list of mods*/
	int found_type; 

	Debug(LDAP_DEBUG_TRACE, "===> xps_create_entries_for_modify: \n",0 , 0, 0);
	
	while ( *modtail != NULL )
	{
		struct xps_x509_type *xps_tail;
		int i;
		found_type = 0;
		Modifications *tmp = ( *modtail );
		
		xps_tail = pkcTypes;
		while ( xps_tail != NULL && !found_type )
		{
			for ( i = 0; xps_tail->attributeTypes[i] != NULL; i++ )
			{
				if ( strcasecmp( tmp->sml_type.bv_val, xps_tail->attributeTypes[i] ) == 0 ) {
					rs->sr_err = xps_process_modify( op, rs, tmp,
						xps_tail->oc, "x509PKC", DN_TYPE_CERT, dn,
						decodeCertificate, "Certificate" , xps_e_array, NULL );
										
					found_type = 1;
					x509_found = 1;
					break;
				}
			}
			xps_tail = xps_tail->next;
		}

		xps_tail = crlTypes;
		while ( xps_tail != NULL && !found_type )
		{
			for ( i = 0; xps_tail->attributeTypes[i] != NULL; i++ )
			{
				if ( strcasecmp( tmp->sml_type.bv_val, xps_tail->attributeTypes[i] ) == 0 ) {
					rs->sr_err = xps_process_modify( op, rs, tmp,
						xps_tail->oc, "x509CRL", DN_TYPE_CRL, dn,
						decodeCertificateList, "certificateList" , xps_e_array, xps_e_revoked );
					found_type = 1;
					x509_found = 1;
					break;
				}
			}
			xps_tail = xps_tail->next;
		}
		
		xps_tail = acTypes;
		while ( xps_tail != NULL && !found_type )
		{
			for ( i = 0; xps_tail->attributeTypes[i] != NULL; i++ )
			{
				if ( strcasecmp( tmp->sml_type.bv_val, xps_tail->attributeTypes[i] ) == 0 ) {
					rs->sr_err = xps_process_modify( op, rs, tmp,
						xps_tail->oc, "x509AC", DN_TYPE_CERT, dn,
						decodeAttributeCertificate, "AttributeCertificate" , xps_e_array, NULL);
					found_type = 1;
					x509_found = 1;
					break;
				}
			}
			xps_tail = xps_tail->next;
		}

		if ( !duplicateAttribute && found_type )
		{
		/* if the X.509 attribute is found and duplicateAttribute is false
			- we don't keep this attr */
			(*modtail) = (*modtail)->sml_next;

			slap_mod_free( &tmp->sml_mod, 0 );
			free( tmp );
		} else {
			if ( (*modtail) != NULL )
				modtail = &(*modtail)->sml_next;
		}
	}
	Debug(LDAP_DEBUG_TRACE, "<=== xps_creat_enries_for_modify: x509_found %d\n",x509_found , 0, 0);
	
	return x509_found;
}

int xps_process_modify (
	Operation *op,
	SlapReply *rs,
	Modifications	*mod,
	char *oc_type,
	char *abs_oc_type,
	int dnflag,
	struct berval dn,
	_CPFPLUS decodeFUNCTION,
	char *decodeName,
	Entry ***e,
	Entry ***e_revoked )
{
	/* Search options */
	Filter *xps_f = NULL;
	struct berval	xps_fstr = { 0, NULL };
	int j;

	Operation xps_op = {0};
	SlapReply xps_rs = {REP_RESULT};
	slap_callback cb = { NULL, xps_send_search_entry, NULL, NULL };

	int rc = LDAP_SUCCESS;
	Debug(LDAP_DEBUG_TRACE, "===> xps_process_modify: \n",0 , 0, 0);
	if ( xps_init( &xps_op, op ) != LDAP_SUCCESS )
		return LDAP_UNWILLING_TO_PERFORM;
	
	switch( mod->sml_op ) {
		case LDAP_MOD_ADD:
			Debug(LDAP_DEBUG_TRACE, "===> xps_process_modify: LDAP_MOD_ADD \n",0 , 0, 0);
			rs->sr_err = xps_x509attrs2entries( op, rs, mod,
				oc_type, abs_oc_type, dnflag, dn,
				decodeFUNCTION, decodeName, e, e_revoked );
			Debug(LDAP_DEBUG_TRACE, "<=== xps_process_modify: \n",0 , 0, 0);
			return rs->sr_err;
		
		case LDAP_MOD_DELETE:
			Debug(LDAP_DEBUG_TRACE, "===> xps_process_modify: LDAP_MOD_DELETE \n",0 , 0, 0);
			if ( mod->sml_values == NULL )
			{ /* no values, i.e. delete all */
				xps_op.o_tag = LDAP_REQ_SEARCH;
				xps_op.o_callback = &cb;
	
				xps_op.ors_scope = LDAP_SCOPE_ONELEVEL;
				xps_op.ors_deref = LDAP_DEREF_NEVER;
				xps_op.ors_slimit = -1;
				xps_op.ors_tlimit = -1;
				xps_op.ors_limit = NULL;
	
				xps_create_filter( op,  &xps_f, &mod->sml_type, NULL );
				xps_op.ors_filter = xps_f;

				filter2bv( xps_f, &xps_fstr );
				xps_op.ors_filterstr =  xps_fstr;
	
				xps_op.o_bd->be_search( &xps_op, &xps_rs );
	
				if ( xps_fstr.bv_val != NULL) 
					free( xps_fstr.bv_val );
				if ( xps_f != NULL)
					filter_free( xps_f );
				
				if ( rc != LDAP_SUCCESS )
					return LDAP_UNWILLING_TO_PERFORM;
				
			} else {
				for ( j = 0; mod->sml_values[j].bv_val != NULL; j++ )
				{
					Modifications	*xps_modlist = NULL;
					Modifications	**xps_modtail = &xps_modlist;
					byte *P, *P_CRL =NULL;
					char *res;
					int len;
	
					struct berval type1 = BER_BVC( "x509serialNumber" );
					struct berval str1;
					struct berval type2 = BER_BVC( "x509issuer" );
					struct berval str2;
					struct berval type3 = BER_BVC( "x509crlThisUpdate" );
					
					P = ( byte *)mod->sml_values[j].bv_val;
					res = decodeFUNCTION(decodeName,0,0,&P,0,0,0,0, NULL, &xps_modtail, &P_CRL);

					P_CRL = NULL;
			
					if ( dnflag == DN_TYPE_CRL )
					{
						str1.bv_val =  xps_find_attr( type3.bv_val, xps_modlist );
						if ( str1.bv_val == NULL )
							return LDAP_UNWILLING_TO_PERFORM;
						str1.bv_len = strlen( str1.bv_val );
						rc = xps_create_filter( op,  &xps_f, &type3, &str1 );
					} else {
						str1.bv_val =  xps_find_attr( type1.bv_val, xps_modlist );
						str1.bv_len = strlen( str1.bv_val );
	
						str2.bv_val =  xps_find_attr( type2.bv_val, xps_modlist );
						str2.bv_len = strlen( str2.bv_val );
						
						xps_f = op->o_tmpalloc( sizeof( Filter ), op->o_tmpmemctx );
						xps_f->f_choice = LDAP_FILTER_AND;
						rc = xps_create_filter( op,  &(xps_f->f_and), &type2, &str2 );
						rc = xps_create_filter( op,  &((xps_f->f_and)->f_next), &type1, &str1 );
						xps_f->f_next =NULL;
					}

					if ( xps_modlist != NULL )
						slap_mods_free( xps_modlist );
					xps_modlist = NULL;
					xps_modtail = &xps_modlist;
			
					
					xps_op.o_tag = LDAP_REQ_SEARCH;
					xps_op.o_callback = &cb;
		
					xps_op.ors_scope = LDAP_SCOPE_ONELEVEL;
					xps_op.ors_deref = LDAP_DEREF_NEVER;
					xps_op.ors_slimit = -1;
					xps_op.ors_tlimit = -1;
					xps_op.ors_limit = NULL;
		
					if ( rc != LDAP_SUCCESS )
						return rc;
					xps_op.ors_filter = xps_f;

					filter2bv( xps_f, &xps_fstr );
					xps_op.ors_filterstr =  xps_fstr;
		
					xps_op.o_bd->be_search( &xps_op, &xps_rs );
	
					if ( xps_fstr.bv_val != NULL) 
						free( xps_fstr.bv_val );
					if ( xps_f != NULL)
						filter_free( xps_f );
					if (str1.bv_val != NULL )
						free( str1.bv_val );
					if (str2.bv_val != NULL )
						free( str2.bv_val );
					
					if ( rc != LDAP_SUCCESS )
						return LDAP_UNWILLING_TO_PERFORM;
				}
			}
			Debug(LDAP_DEBUG_TRACE, "<=== xps_process_modify: \n",0 , 0, 0);
			return rs->sr_err;
		
		case LDAP_MOD_REPLACE:
				xps_op.o_tag = LDAP_REQ_SEARCH;
				xps_op.o_callback = &cb;
	
				xps_op.ors_scope = LDAP_SCOPE_ONELEVEL;
				xps_op.ors_deref = LDAP_DEREF_NEVER;
				xps_op.ors_slimit = -1;
				xps_op.ors_tlimit = -1;
				xps_op.ors_limit = NULL;
	
				xps_create_filter( op,  &xps_f, &mod->sml_type, NULL );
				xps_op.ors_filter = xps_f;

				filter2bv( xps_f, &xps_fstr );
				xps_op.ors_filterstr =  xps_fstr;
	
				xps_op.o_bd->be_search( &xps_op, &xps_rs );
	
				if ( xps_fstr.bv_val != NULL) 
					free( xps_fstr.bv_val );
				if ( xps_f != NULL)
					filter_free( xps_f );
				
				if ( rc != LDAP_SUCCESS )
					return LDAP_UNWILLING_TO_PERFORM;

				if ( mod->sml_values != NULL )
				{
					rs->sr_err = xps_x509attrs2entries( op, rs, mod,
						oc_type, abs_oc_type, dnflag, dn,
						decodeFUNCTION, decodeName, e, e_revoked );
				} else
					rs->sr_err = LDAP_UNWILLING_TO_PERFORM;

			Debug(LDAP_DEBUG_TRACE, "<=== xps_process_modify: \n",0 , 0, 0);
			return rs->sr_err;

		default:
			Debug(LDAP_DEBUG_TRACE, "<=== xps_process_modify: \n",0 , 0, 0);
			return LDAP_UNWILLING_TO_PERFORM;
	}
}

int xps_get_level3_entries( Operation *xps_op,	SlapReply *rs )
{
	int i;
	int rc = LDAP_SUCCESS;
	Filter *xps_f = NULL;
	struct berval	xps_fstr = { 0, NULL };
	struct berval	type = { 11, "objectClass" };
	slap_callback *cb1, cb = { NULL, xps_send_search_entry, NULL, NULL };

	AttributeDescription *ad;
	const char *text;
	Attribute *a;
	BerVarray bv;

	Debug(LDAP_DEBUG_TRACE, "===> xps_get_level3_entries \n",0 , 0, 0);

	if ( xps_e_level2 == NULL ) {
		return LDAP_SUCCESS;
	}
	xps_op->o_tag = LDAP_REQ_SEARCH;
	xps_op->ors_scope = LDAP_SCOPE_ONELEVEL;
	xps_op->ors_deref = LDAP_DEREF_NEVER;
	xps_op->ors_slimit = -1;
	xps_op->ors_tlimit = -1;
	xps_op->ors_limit = NULL;
	cb1 = xps_op->o_callback;
	xps_op->o_callback = &cb;
	xps_create_filter( xps_op,  &xps_f, &type, NULL );
	xps_op->ors_filter = xps_f;
	filter2bv( xps_f, &xps_fstr );
	xps_op->ors_filterstr =  xps_fstr;
	
	for ( i = 0; xps_e_level2[i] != NULL; i++ )
	{

		ad = NULL;
		slap_str2ad( "objectClass", &ad, &text );
	
		a = attr_find( xps_e_level2[i]->e_attrs, ad );
	
		if ( a->a_vals != NULL )
		{
			for ( bv = a->a_vals; bv->bv_val; bv++ ) {
				if ( strcasecmp( bv->bv_val, "x509CRL" ) ) {
							
					xps_op->o_req_dn = xps_e_level2[i]->e_name;
					xps_op->o_req_ndn = xps_e_level2[i]->e_nname;
					xps_op->o_bd->be_search( xps_op, rs );
					break;
				}
			}
		}
	}
	xps_op->o_callback = cb1;
	Debug(LDAP_DEBUG_TRACE, "<=== xps_get_level3_entries: \n",0 , 0, 0);
}
